/*
 * This program is original work by San Mehat <nettwerk@valinux.com> and
 * Zac Sprackett <zacs@valinux.com>.
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 2001, San Mehat, Zac Sprackett, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <libloose.h>
#include <quanta_module_harness.h>
#include <prototypes.h>

static SLAVE		*slave_head = NULL;
static SLAVE		*slave_tail = NULL;
extern NODELIST	*nodelist_head;
extern NODELIST	*nodelist_tail;
extern int num_nodes_per_slave;

static SLAVE	*slave_create(void);
static SLAVE	*slave_alloc(void);
static void		slave_free(SLAVE *slave);
static void		slave_msg_handler(int fd, void *data);
static void		free_parent_resources(SLAVE *slave);

/*
 * Assign a node to a slave and notify the slave
 */
int slave_assign_and_notify(NODELIST *node)
{
	if (!(node->slave = locate_available_slave()))
		{
		if (!(node->slave = slave_create()))
			return(-ENOMEM);
		}
	node->slave->nodecount++;
	return(slave_notify_new_node(node));
}

/*
 * Notify a slave that a node is disapearing
 * Kill the slave if it has no nodes
 */
int slave_remove_node(NODELIST *node)
{
	int	error;

	node->slave->nodecount--;
	error = slave_notify_del_node(node);
	if (node->slave->nodecount == 0)
		slave_destroy(node->slave);
	node->slave = NULL;
	return(error);
}

/*
 * Find a slave with some free slots in it
 */
SLAVE *locate_available_slave()
{
	SLAVE	*slave_ptr;

	slave_ptr = slave_head;
	while(slave_ptr)
		{
		if (slave_ptr->nodecount < num_nodes_per_slave)
			return(slave_ptr);
		slave_ptr = slave_ptr->next;
		}
	return(NULL);
}

/*
 * Notify a slave of a variable change 
 */
int slave_notify_variable(NODELIST *node, char *variable, char *value)
{
	char	msg[255];
	int		error;

	if (!node->slave)
		return(-EBADSLT);
	sprintf(msg, "NODE_VAR:%s:%s:%s", node->node_id, variable, value);
	if ((error = slave_send(node->slave, msg)) < 0)
		return(error);
	return(0);
}

/*
 * Notify a slave that a client has disconnected
 */
int slave_notify_disconnection(int client_fd)
{
	char	msg[255];
	SLAVE	*slv;
	int		error;

	sprintf(msg,"CLIENT_DIS:NONE:%d",client_fd);
	slv = slave_head;
	while(slv)
		{
		if ((error = slave_send(slv, msg)) < 0)
			return(error);
		slv = slv->next;
		}
	return(0);
}

/*
 * Notify a slave that a node name has changed
 */
int slave_notify_rename_node(NODELIST *node, char *new_name)
{
	char	msg[255];
	int		error;

	sprintf(msg, "NODE_RENAME:%s:%s", node->node_id, new_name);
	if ((error = slave_send(node->slave, msg)) < 0)
		return(error);
	return(0);
}

/*
 * Notify a slave that a node is gone (deleted)
 */
int slave_notify_del_node(NODELIST *node)
{
	char	msg[255];
	int		error;

	if (!node->slave)
		return(-EBADSLT);
	sprintf(msg, "NODE_DEL:%s", node->node_id);
	if ((error = slave_send(node->slave, msg)) < 0)
		return(error);
	return(0);
}

/*
 * Notify a slave of its new burden (node)
 * Returns 0 on success, negative errno on failure
 */
int slave_notify_new_node(NODELIST *node)
{
	char	msg[255];
	int		error;

	if (!node->slave)
		return(-EBADSLT);
	sprintf(msg, "NODE_NEW:%s", node->node_id);
	if ((error = slave_send(node->slave, msg)) < 0)
		return(error);
	return(0);
}

/*
 * Notify all slaves of shutdown
 */
int slave_deinit()
{
	SLAVE	*slv;
	char	slave_msg[255];

	strcpy(slave_msg,"SLAVE_SHUTDOWN:NONE:NONE");
	slv = slave_head;
	while(slv)
		{
		slave_send(slv, slave_msg);
		slv = slv -> next;
		}
	return(0);
}

/*
 * Allocate a slave structure and fork off a slave
 */
static SLAVE *slave_create()
{
	SLAVE	*new_slave;
	int		pipefd1[2];
	int		pipefd2[2];

	if ((pipe(pipefd1) < 0 ) || (pipe(pipefd2) < 0))
		{
		lm_log("Unable to create thread pipes");
		return(NULL);
		}
	if (!(new_slave = slave_alloc()))
		return(NULL);
	new_slave->read_sock = pipefd1[0];
	new_slave->write_sock = pipefd2[1];
	if (!(new_slave->pid = fork()))
		{
		/* New slave */
		free_parent_resources(new_slave);
		slave_machine_start(pipefd2[0], pipefd1[1]);
		_exit(0);
		}
	new_slave->nodecount = 0;
	lm_watch_fd(new_slave->read_sock, slave_msg_handler, (void *) new_slave);
	return(new_slave);
}

/*
 * Called from within the child context
 * Frees up memory and closes file descriptors
 * that the child wont use (that were
 * inherited from the parent
 */
static void free_parent_resources(SLAVE *slave)
{
	SLAVE			*slave_scan;
	NODELIST	*node_scan;

	slave_scan = slave_head;
	while(slave_scan)
		{
		if (slave_scan != slave)
			{
			/* We don't want to close our own read/write fd's */
			close(slave_scan->read_sock);
			close(slave_scan->write_sock);
			}
		slave_free(slave_scan);
		slave_scan = slave_head;
		}

	node_scan = nodelist_head;
	while(node_scan)
		{
		node_destroy(node_scan);
		node_scan = nodelist_head;
		}
}

/*
 * Shut down a slave and free up it's structure
 */
void slave_destroy(SLAVE *slave)
{
	NODELIST	*nodes;
	char			msg[255];

	/* Ask the slave to shutdown cleanly */
	strcpy(msg,"SLAVE_SHUTDOWN:NONE:NONE");
	slave_send(slave, msg);
	sleep(1);

	/* Send the slave a signal */
	kill(slave->pid, SIGKILL);

	/* Clean up any nodes that were using this slave */
	nodes = nodelist_head;
	while(nodes)
		{
		if (nodes->slave == slave)
			nodes->slave = NULL;
		nodes = nodes->next;
		}

	/* Now free the resources */
	lm_unwatch_fd(slave->read_sock);
	close(slave->read_sock);
	close(slave->write_sock);
	slave_free(slave);
}

/*
 * Allocate a slave structure
 */
static SLAVE *slave_alloc()
{
	SLAVE	*new;

	if (!(new = (SLAVE *) malloc(sizeof(SLAVE))))
		{
		lm_log("Unable to malloc new slave structure");
		return(NULL);
		}
	memset(new, 0, sizeof(SLAVE));

  if (!slave_head)
    {
    slave_head = new;
    slave_tail = new;
    }
  else
    {
    slave_tail->next = new;
    new->prev = slave_tail;
    slave_tail = new;
    }
	return(new);
}

/*
 * Free a slave structure
 */
static void slave_free(SLAVE *slave)
{
  if (slave == slave_head)
    {
    slave_head = slave_head->next;
    if (slave_head)
      slave_head->prev = NULL;
		else
			slave_tail = NULL;
    }
  else if (slave == slave_tail)
    {
    slave_tail = slave_tail->prev;
    if (slave_tail)
      slave_tail->next = NULL;
		else
			slave_head = NULL;
    }
  else
    {
    slave->next->prev = slave->prev;
    slave->prev->next = slave->next;
    }
	free(slave);
}

int slave_send(SLAVE	*slave, char *msg)
{
	__uint32_t	len;
	__uint32_t	net_len;
	int					rc;

#if 0
	printf("Sending msg [%s] to slave pid %d (fd %d)\n",
					msg,
					slave->pid,
					slave->write_sock);
#endif

	len = strlen(msg)+1;
	net_len = htonl(len);
	if ((rc = write (slave->write_sock, (void *) &net_len, sizeof(net_len))) < 0)
		return(rc);
	if ((rc = write (slave->write_sock, (void *) msg, len)) < 0)
		return(rc);
	return(0);
}

static int slave_recv(int fd, char **buffer, __uint32_t *len)
{
  __uint32_t  net_len;
	__uint32_t	brtr;
  int         rc;
	char				*offset;

  if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
  *len = ntohl(net_len);

  if (!(*buffer = malloc(*len)))
    return(-ENOMEM);
  memset(*buffer, 0, *len);
	brtr = *len;
	offset =*buffer;
	while(brtr)
		{
  	if ((rc = read(fd, (void *) offset, brtr)) < 0)
    	return(rc);
		brtr -= rc;
		offset += rc;
		}
  return(0);
}

/*
 * Handler for messages comming from a slave to the master 
 */
static void slave_msg_handler(int fd, void *data)
{
	int	rc, len;
	char	*msg_buffer;

	KNOWN_UNUSED_PARAM(data);

	if ((rc = slave_recv(fd, &msg_buffer, &len)) < 0)
		{
		printf("Master unable to read msg from slave (%s)\n",strerror(-rc));
		return;
		}

	lm_send_to_nexxus(msg_buffer);

	free(msg_buffer);
	return;
}
