
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> and
 * Zac Sprackett <zacs@valinux.com>.
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 2001, San Mehat, Zac Sprackett, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <netdb.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <quanta_module_harness.h>
#include <quanta_slave_context.h>
#include <prototypes.h>
#include <quanta_slave_context.h>
#include <pthread.h>
#include <quanta.h>

static void	thread_cleanup(void *);
static void	process_msg_from_slave(CONTEXT *);
static int	recv_msg(int , char **, __uint32_t *);
static void	process_quanta_ipc(CONTEXT *, __uint32_t, int, char *, char *);
static int	send_msg(int , char *);
static int	quanta_engage(CONTEXT *);
static void	quanta_disengage(CONTEXT *);
static int  quanta_reengage(CONTEXT *);
static void set_device_address(CONTEXT *, char *);

/* QUANTA forward declarations */
static int init_quanta(CONTEXT *);
static int do_power_on(CONTEXT *);
static int do_power_off(CONTEXT *);
static int do_reset(CONTEXT *);
static void flush_buffer(CONTEXT *ctx);
static int wait_for_data(CONTEXT *ctx);
static int	do_quanta_transaction(CONTEXT *, unsigned char *, int,
			unsigned char *, int *);
static void process_quanta_ipc(CONTEXT *, __uint32_t, int, char *, char *);
static int	do_get_fw_version(CONTEXT *ctx, FW_VER *ver);
static int	do_misc_led_on(CONTEXT *ctx);
static int	do_misc_led_off(CONTEXT *ctx);
static int	do_misc_led_toggle(CONTEXT *ctx);
static int	do_get_bmc_status(CONTEXT *ctx, BMC_STAT *stat);
static int  do_ping(CONTEXT *ctx);
static int	do_get_sensor_limits(CONTEXT *ctx, C_SENSOR_LIM *c_lim);
static int	do_get_monitoring_data(CONTEXT *ctx, C_MD *c_md);

/*
 * This stuff is so that a thread can 'peek' at another threads context
 * structure (to check for duplicate serial ports, etc)
 */
extern pthread_mutex_t	context_mutex;
extern CONTEXT					*context_head;

static void thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	
	quanta_disengage(ctx);
	ctx->refcount--;
}

void *thread_start(void *data)
{
	struct _pthread_cleanup_buffer 	_buff;
	CONTEXT													*ctx = (CONTEXT *) data;
	struct timeval									to;
	fd_set													read_fds;
	int															max, rc;
	char														log_msg[255];

#ifdef THREAD_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s starting (pid %d)",
					ctx->node_id,
					getpid());
	send_msg(ctx->thread_ipc_write, log_msg);
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	_pthread_cleanup_push(&_buff, thread_cleanup, (void *) ctx);
	ctx->refcount++;

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(ctx->thread_ipc_read, &read_fds);
		max = ctx->thread_ipc_read;

		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s select() failed (%s)",
							ctx->node_id,
							strerror(errno));
			send_msg(ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
			}
		if (!rc)
			{
			pthread_testcancel();
			continue;
			}
		if (FD_ISSET(ctx->thread_ipc_read, &read_fds))
			{
			process_msg_from_slave(ctx);
			}
		pthread_testcancel();
		}
}

static void process_msg_from_slave(CONTEXT *ctx)
{
	char				*msg_buffer, *cmd, *node_id, *bword, *saved_buffer;
	char				*ipc_cmd, *ipc_args, *variable, *value;
	int					len, rc, client_fd;
	__uint32_t	job_id;
	char				*last = (char *) 0x00;
	char				log_msg[255];

	if ((rc = recv_msg(ctx->thread_ipc_read, &msg_buffer, &len)) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to read msg from slave (%s)",
					 	ctx->node_id,
					 	strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		return;
		}
	saved_buffer = malloc(len);
	memcpy(saved_buffer, msg_buffer, len);

	if (!(cmd = strtok_r(msg_buffer, ":", &last)))
		goto out;
	if (!(node_id = strtok_r(NULL, ":", &last)))
		goto out;
	if (!strcasecmp(cmd,"NODE_IPC"))
		{
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(bword);
		if (!(ipc_cmd = strtok_r(NULL, ":", &last)))
			goto out;
		ipc_args = &saved_buffer[((ipc_cmd - msg_buffer)+strlen(ipc_cmd)+1)];
		process_quanta_ipc(ctx, job_id, client_fd, ipc_cmd, ipc_args);
		}
	else if (!strcasecmp(cmd, "NODE_VAR"))
		{
		if (!(variable = strtok_r(NULL, ":", &last)))
			goto out;

		/* BEGIN San mod 07/12/00 */
		value = &saved_buffer[((variable - msg_buffer)+(strlen(variable)+1))];
		/* END San mod 07/12/00 */
		if (!strcasecmp(variable,"DEVICE_ADDRESS"))
			{
			set_device_address(ctx, value);
			quanta_disengage(ctx);

			/* If we have a device, we should engage */
			if (!strcasecmp(value, "NONE"))
				quanta_disengage(ctx);
			else
				{
				if ((rc = quanta_engage(ctx)) < 0)
					{
					return;
					}
				}
			goto out;
			}
		}
	else if (!strcasecmp(cmd, "CLIENT_DIS"))
		{
		char								*fd;
		int									client_fd;

		/* Called when a client disconnects from nexxus */
		if (!(fd = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(fd);
		goto out;
		}
	else if (!strcasecmp(cmd, "NODE_RENAME"))
		{
		char	*new_name;
		char  *tmp;
	
		/* Called when a node is renamed */
		if (!(new_name = strtok_r(NULL, ":", &last)))
			goto out;

		tmp = ctx->node_id;
		ctx->node_id = (char *) malloc(strlen(new_name)+1);
		strcpy(ctx->node_id, new_name);
		free(tmp);
		goto out;
		}
	else
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s received unsupported cmd from slave (%s)",
					 	ctx->node_id,
					 	cmd);
		send_msg(ctx->thread_ipc_write, log_msg);
		}

out:
	free(saved_buffer);
	free(msg_buffer);
}

/*
 * Process an IPC message
 */
static void process_quanta_ipc(CONTEXT			*ctx,
															__uint32_t	job_id,
															int					client_fd,
															char				*ipc_cmd,
															char				*ipc_args)
{
	char					resp_msg[255];
	char					*last = (char *) 0x00;
	int						rc;

	if (!strcasecmp(ipc_cmd, "CONFIGURATION"))
		{
		CONTEXT *ctx_scan;
		char *dev;

		if (!(dev = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (strcasecmp(dev,"NONE"))
			{
			/* Scan for device in use */
			pthread_mutex_lock(&context_mutex);
			ctx_scan = context_head;
			while(ctx_scan)
				{
				if (ctx_scan->device_address)
					if ((!strcmp(ctx_scan->device_address,dev)) &&
						(ctx_scan != ctx))
						break;
					ctx_scan = ctx_scan->next;
				}
			pthread_mutex_unlock(&context_mutex);
			if (ctx_scan)
				{
				sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR,%s",
					job_id,
					client_fd,
					strerror(EADDRINUSE));
				send_msg(ctx->thread_ipc_write, resp_msg);
				goto out_ok;
				}
			}

		/* See if the serial port has changed */
		if ((ctx->device_address ? strcmp(ctx->device_address, dev) : 1))
			{
			sprintf(resp_msg,
				"0:0:SET:%s:QUANTA:DEVICE_ADDRESS:%s",
				ctx->node_id,
				dev);
			send_msg(ctx->thread_ipc_write,resp_msg);
			set_device_address(ctx,dev);

			/* Reengage on the new serial port */
			quanta_reengage(ctx);
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	/* Anything below this requires a configured device */
	if (!ctx->device_address)
		{
		sprintf(resp_msg,
			"%d:%d:FOR:JOB_ERROR:%s",
		job_id, client_fd, strerror(ENOPKG));
		send_msg(ctx->thread_ipc_write,resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "REFRESH"))
		{
		quanta_reengage(ctx);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "INQUIRY"))
		{
		sprintf(resp_msg,
			"%d:%d:FOR:INQUIRY:%s:%s",
			job_id,client_fd,ctx->node_id,ctx->device_address);
		send_msg(ctx->thread_ipc_write,resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}

	/* All of these require a connected device */
	if (ctx->con_sock == -1)
		{
		sprintf(resp_msg,
			"%d:%d:FOR:JOB_ERROR:%s",
		job_id, client_fd, strerror(ENOTCONN));
		send_msg(ctx->thread_ipc_write,resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "POWER_ON"))
		{
		if (do_power_on(ctx) == -1)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(EPROTO));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "POWER_OFF"))
		{
		if (do_power_off(ctx) == -1)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(EPROTO));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "HARD_RESET"))
		{
		if (do_reset(ctx) == -1)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(EPROTO));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "IDENTIFY"))
		{
		int i;

		for (i=0; i<3; i++)
			{
			if (do_misc_led_on(ctx) == -1)
				{
				sprintf(resp_msg,
					"%d:%d:FOR:JOB_ERROR:%s",
					job_id,client_fd,strerror(EPROTO));
				send_msg(ctx->thread_ipc_write,resp_msg);
				goto out_ok;
				}
			sleep(1); 
			if ((rc = do_misc_led_off(ctx))<0)
				{
				sprintf(resp_msg,
					"%d:%d:FOR:JOB_ERROR:%s",
					job_id,client_fd,strerror(-rc));
				send_msg(ctx->thread_ipc_write,resp_msg);
				goto out_ok;
				}
			sleep(1); 
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "READ_ANALOG_SENSORS"))
		{
		C_MD          c_md;

		if ((rc = do_get_monitoring_data(ctx,&c_md))<0)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(EPROTO));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, 
			"%d:%d:FOR:READ_ANALOG_SENSORS:%.3f:%.3f:%.3f:%.3f:%.3f:%.3f:%.3f:%.3f:%.3f:%.3f:%.1f:%.1f:%.1f:%.1f:%.1f:%.3f:%.3f",
			job_id,
			client_fd,
			c_md.cpu2_vcore,
			c_md.mainboard_plus_2_7,
			c_md.pwr_plus_3_3,
			c_md.pwr_plus_12,
			c_md.pwr_minus_12,
			c_md.pwr_plus_5,
			c_md.mainboard_plus_1_5,
			c_md.cpu1_vcore,
			c_md.stndby_plus_3_3,
			c_md.scsi_plus_2_85,
			c_md.sys_temp,
			c_md.cpu2_temp,
			c_md.cpu1_temp,
			c_md.fan2_speed,
			c_md.fan1_speed,
			c_md.batt_volt_plus_3_3,
			c_md.stndby_plus_5);
		send_msg(ctx->thread_ipc_write,resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "READ_ANALOG_LIMITS"))
		{
		C_SENSOR_LIM          c_lim;

		if ((rc = do_get_sensor_limits(ctx,&c_lim))<0)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(EPROTO));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, 
			"%d:%d:FOR:READ_ANALOG_LIMITS:%.3f/%.3f:%.3f/%.3f:%.3f/%.3f:%.3f/%.3f:%.3f/%.3f:%.3f/%.3f:%.3f/%.3f:%.3f/%.3f:%.3f/%.3f:%.3f/%.3f:%.1f:%.1f:%.1f:%.1f:%.1f:%.3f/%.3f:%.3f/%.3f",
			job_id,
			client_fd,
			c_lim.cpu2_vcore_l,          c_lim.cpu2_vcore_h,
			c_lim.mainboard_plus_2_7_l,  c_lim.mainboard_plus_2_7_h,
			c_lim.pwr_plus_3_3_l,        c_lim.pwr_plus_3_3_h,
			c_lim.pwr_plus_12_l,         c_lim.pwr_plus_12_h,
			c_lim.pwr_minus_12_l,        c_lim.pwr_minus_12_h,
			c_lim.pwr_plus_5_l,          c_lim.pwr_plus_5_h,
			c_lim.mainboard_plus_1_5_l,  c_lim.mainboard_plus_1_5_h,
			c_lim.cpu1_vcore_l,          c_lim.cpu1_vcore_h,
			c_lim.stndby_plus_3_3_l,     c_lim.stndby_plus_3_3_h,
			c_lim.scsi_plus_2_85_l,      c_lim.scsi_plus_2_85_h,
			c_lim.sys_temp_h,
			c_lim.cpu2_temp_h,
			c_lim.cpu1_temp_h,
			c_lim.fan1_speed_l,
			c_lim.fan2_speed_l,
			c_lim.batt_volt_plus_3_3_l,  c_lim.batt_volt_plus_3_3_h,
			c_lim.stndby_plus_5_l,       c_lim.stndby_plus_5_h);
		send_msg(ctx->thread_ipc_write,resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "READ_DIGITAL_SENSORS"))
		{
		BMC_STAT         stat;

		if ((rc = do_get_bmc_status(ctx,&stat))<0)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(EPROTO));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, 
			"%d:%d:FOR:READ_DIGITAL_SENSORS:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d",
			job_id,
			stat.s1.cpu2_vcore,
			stat.s1.two_pt_five,
			stat.s1.three_pt_three,
			stat.s1.five,
			stat.s1.sys_temp,
			stat.s1.cpu2_temp,
			stat.s1.fan1,
			stat.s1.fan2,
			stat.s2.twelve,
			stat.s2.minus_twelve,
			stat.s2.three_pt_three_stndby,
			stat.s2.cpu1_temp,
			stat.s2.cpu1_vcore,
			stat.s2.five_stndby,
			stat.sb.mem_ecc_1bit_err,
			stat.sb.mem_ecc_multibit_err,
			stat.sb.pci_perr,
			stat.sb.pci_serr,stat.s3.mb_one_pt_five,
			stat.s3.scsi_two_pt_eightfive,
			stat.s3.batt_three_pt_three,
			stat.s3.chassis_intrusion,
			stat.s3.cpu1_ierr,
			stat.s3.cpu2_ierr,
			stat.s3.cpu1_thermal_trip,
			stat.s3.cpu2_thermal_trip);
		send_msg(ctx->thread_ipc_write,resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "CHASSIS_STATUS"))
		{
		C_MD          c_md;

		if ((rc = do_get_monitoring_data(ctx,&c_md))<0)
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(EPROTO));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		if (c_md.pwr_plus_3_3) 
			{
			sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:POWERED_ON",job_id,client_fd);
			send_msg(ctx->thread_ipc_write, resp_msg);
			}
		else
			{
			sprintf(resp_msg, "%d:%d:FOR:CHASSIS_STATUS:POWERED_OFF",job_id,client_fd);
			send_msg(ctx->thread_ipc_write, resp_msg);
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "BMC_INFO"))
		{
		FW_VER ver;

		if (do_get_fw_version(ctx,&ver))
			{
			sprintf(resp_msg,
				"%d:%d:FOR:JOB_ERROR:%s",
				job_id,client_fd,strerror(EINTR));
			send_msg(ctx->thread_ipc_write,resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg,
			"%d:%d:FOR:BMC_INFO:%d.%d",
			job_id,client_fd,ver.major,ver.minor);
		send_msg(ctx->thread_ipc_write, resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(EBADRQC));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}

out_ok:
	return;

out_malformatted:
	sprintf(resp_msg,
					"%d:%d:FOR:JOB_ERROR:%s",
					job_id,
					client_fd,
					strerror(EINVAL));
	send_msg(ctx->thread_ipc_write, resp_msg);
	return;
}
	
/*
 * Receive a message. This function is threadsafe
 */
static int recv_msg(int fd, char **buffer, __uint32_t *len)
{
  __uint32_t  net_len;
	__uint32_t	brtr;
	char				*offset;
  int         rc;

  if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
	else if (!rc)
		return(-1);
  *len = ntohl(net_len);

  if (!(*buffer = malloc(*len)))
    return(-ENOMEM);
  memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr)
		{
  	if ((rc = read(fd, (void *) offset, brtr)) < 0)
    	return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
		}
  return(0);
}

/*
 * Send a message. This function is threadsafe
 */
static int send_msg(int fd, char *msg)
{
  __uint32_t  len;
  __uint32_t  net_len;
  int         rc;

	if (fd == -1)
		return(-EBADF);
  len = strlen(msg)+1;
  net_len = htonl(len);
  if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
  if ((rc = write (fd, (void *) msg, len)) < 0)
    return(rc);
  return(0);
}

static int quanta_engage(CONTEXT *ctx)
{
	char *last = (char *) 0x00;
	char log_msg[255];
#ifndef __sgi
	struct serial_struct  ser;
#endif /* __sgi */

	if (ctx->con_sock != -1)
		return(-EADDRINUSE);

	if (!ctx->device_address)
		return(-EFAULT);

	/* Is this a serial port or a network address */
	if (ctx->device_address[0] == '/')
		{
		struct termios tty;
		if ((ctx->con_sock = open(ctx->device_address, O_RDWR))<0)
			{
			sprintf(log_msg,"0:0:LOG:Thread %s failed to engage on serial port (%m)",
				ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
		tcgetattr(ctx->con_sock, &tty);
		cfmakeraw(&tty);
		cfsetospeed(&tty,B19200);
		cfsetispeed(&tty,B19200);
		tcsetattr(ctx->con_sock,TCSANOW,&tty);
#ifndef __sgi
		if (ioctl (ctx->con_sock, TIOCGSERIAL, &ser) == 0)
			{
			if (!(ser.flags & ASYNC_LOW_LATENCY))
				{
				ser.flags |= ASYNC_LOW_LATENCY;
				ioctl (ctx->con_sock, TIOCSSERIAL, &ser);
				}
			}
#endif  /*__sgi*/
		}
	else
		{
		struct sockaddr_in sock;
		int optval;
		char address[64];
		unsigned int port;
		char *bword;
		char tmp[64];

		strcpy(tmp, ctx->device_address);

		if (!(bword = strtok_r(tmp, "-", &last)))
			return(-EFAULT);
		strcpy(address,bword);

		if (!(bword = strtok_r(NULL, "-", &last)))
			return(-EFAULT);

		port = atoi(bword);

		if ((ctx->con_sock = socket(AF_INET, SOCK_STREAM, 0))<0)
			{
			sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
		optval = 1;
		if (setsockopt(ctx->con_sock,SOL_SOCKET,SO_REUSEADDR,
			&optval,sizeof(optval))<0)
				{
				sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
				send_msg(ctx->thread_ipc_write, log_msg);
				return(-errno);
				}

		memset(&sock, 0, sizeof(sock));
		sock.sin_family = AF_INET;
		sock.sin_addr.s_addr = inet_addr(address);
		sock.sin_port = htons(port);

		if (connect(ctx->con_sock,(struct sockaddr *) &sock,sizeof(sock)) < 0)
			{
			sprintf(log_msg,"0:0:LOG:Thread %s engage failed (%m)",ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
		}
	sprintf(log_msg,"0:0:LOG:Thread %s engage successful",ctx->node_id);
	send_msg(ctx->thread_ipc_write, log_msg);
	if (init_quanta(ctx)<0)
		{
		quanta_disengage(ctx);
		return(-1);
		}

	return(0);
}

static void quanta_disengage(CONTEXT *ctx)
{
	if (ctx->con_sock != -1)
		{
		char log_msg[255];
		int  tmp_sock;

		sprintf(log_msg, "0:0:LOG:Thread %s disengaging from %s",
			ctx->node_id, ctx->device_address);
		send_msg(ctx->thread_ipc_write, log_msg);

		tmp_sock = ctx->con_sock;
		ctx->con_sock = -1;
		close(tmp_sock);
		}
}

static int quanta_reengage(CONTEXT *ctx)
{
	quanta_disengage(ctx);
	return(quanta_engage(ctx));
}

static void set_device_address(CONTEXT *ctx, char *dev)
{
	if (ctx->device_address)
		free(ctx->device_address);
	ctx->device_address = malloc((strlen(dev)+1));
	strcpy(ctx->device_address,dev);
}

/* QUANTA FUNCTIONS */
static int bps_table[] = {B230400, B115200, B57600, B38400, B19200, B9600, B2400, 0};

static int init_quanta(CONTEXT *ctx)
{
	struct termios	tty;
	int							i, rc;
	char						buffer[1024*64];
	char						line[6];
	char						*p;
	int							brtr;
	

	// Try to auto-baud.. start high.. work down
	for (i=0;bps_table[i] != 0; i++)
		{
  	tcgetattr(ctx->con_sock,&tty);
  	cfmakeraw(&tty);
  	cfsetospeed(&tty,bps_table[i]);
  	cfsetispeed(&tty,bps_table[i]);
  	tcsetattr(ctx->con_sock,TCSANOW,&tty);

		flush_buffer(ctx);
		strcpy(line,"\n\r");
	
		write(ctx->con_sock, &line[0], 2);

		p = &buffer[0];
		brtr = sizeof(buffer);
		if (!(rc = wait_for_data(ctx)))
			continue;

		while (wait_for_data(ctx))
			{
			int len;
			len = read(ctx->con_sock, p, brtr);
			p+=len;
			brtr -=len;
			if (!brtr)
				{
				flush_buffer(ctx);
				break;
				}
			}
		// Check buffer
		if (strstr(buffer, "SU6"))
			return(0);
		}
	return(-1);
}

static int wait_for_data(CONTEXT *ctx)
{
	fd_set					fds;
	struct timeval	to;
	int							rc;

	FD_ZERO(&fds);
	FD_SET(ctx->con_sock, &fds);
	to.tv_usec = 0;
	to.tv_sec = 1;
	if ((rc = select((ctx->con_sock+1), &fds, NULL, NULL, &to))<0)
		{
		perror("select");
		return(-1);
		}
	else if (!rc)
		return(0);
	return(1);
}

static int	do_get_fw_version(CONTEXT *ctx, FW_VER *ver)
{
	unsigned char	req[3];
	int						rc, len;

	req[0] = 0x53;
	req[1] = CMD_GET_FW_VERSION;
	req[2] = req[0] + req[1];
	len = sizeof(FW_VER);
	rc = do_quanta_transaction(ctx, req, 3,(unsigned char *) ver, &len);
	return(0);
}

static int	do_get_bmc_status(CONTEXT *ctx, BMC_STAT *stat)
{
	unsigned char	req[3];
	int						rc, len, i;
	unsigned char	cksum = 0;
	unsigned char	*p;

	req[0] = 0x53;
	req[1] = CMD_GET_BMC_STATUS;
	req[2] = req[0] + req[1];
	len = sizeof(BMC_STAT);
	rc = do_quanta_transaction(ctx, req,3,(unsigned char *) stat, &len);

#if 1
	p = (unsigned char *) stat;
	for (i=0; i < (int)(sizeof(BMC_STAT)-1); i++)
		{
		printf("data[%d] = 0x%.2x (cksum = 0x%.2x)\n",i,p[i], cksum);
		cksum += p[i];
		}
	if (cksum != stat->cksum)
		{
		printf("pkt cksum = 0x%.2x, ours = 0x%.2x\n",stat->cksum, cksum);
		return(-1);
		}
#endif
	return(0);
}

static int	do_get_sensor_limits(CONTEXT *ctx, C_SENSOR_LIM *c_lim)
{
	unsigned char	req[3];
	int						rc, len, i;
	SENSOR_LIM		rep;
	unsigned char	cksum = 0;
	unsigned char	*p;


	memset(c_lim, 0, sizeof(C_SENSOR_LIM));

	while(1)
		{
		req[0] = 0x53;
		req[1] = CMD_GET_SENSOR_LIMITS;
		req[2] = req[0] + req[1];
		len = sizeof(rep);
		rc = do_quanta_transaction(ctx, req, 3, (unsigned char *)&rep, &len);
		if (!len)
			{
			printf("0 len\n");
			continue;
			}
		if (len != sizeof(rep))
			{
			printf("repsize = %d, len = %d\n",sizeof(rep), len);
			return(-1);
			}
		p = (unsigned char *) &rep;
		for (i=0; i < (int)(sizeof(rep)-1); i++)
			cksum += p[i];
		printf("cksum = 0x%.2x, pkt = 0x%.2x\n",cksum, rep.cksum);
		break;
		}

	c_lim->cpu2_vcore_h					=(float) (rep.cpu2_vcore_h *16) / (float) 1000;
	c_lim->cpu2_vcore_l					=(float) (rep.cpu2_vcore_l *16) / (float) 1000;
	c_lim->mainboard_plus_2_7_h	=(float) (rep.mainboard_plus_2_7_h * 16) / (float) 1000;
	c_lim->mainboard_plus_2_7_l	=(float) (rep.mainboard_plus_2_7_l * 16) / (float) 1000;
	c_lim->pwr_plus_3_3_h				=(float) (rep.pwr_plus_3_3_h * 16) / (float) 1000;
	c_lim->pwr_plus_3_3_l				=(float) (rep.pwr_plus_3_3_l * 16) / (float) 1000;
	c_lim->pwr_plus_12_h				=(float) (rep.pwr_plus_12_h * 61) / (float) 1000;
	c_lim->pwr_plus_12_l				=(float) (rep.pwr_plus_12_l * 61) / (float) 1000;
	c_lim->pwr_minus_12_h				=(float) (((rep.pwr_minus_12_h * 16) / ((float)1000 - (float) 2.9016)) / (float) 0.194);
	c_lim->pwr_minus_12_l				=(float) (((rep.pwr_minus_12_l * 16) / ((float)1000 - (float) 2.9016)) / (float) 0.194);
	c_lim->pwr_plus_5_h					=(float) (rep.pwr_plus_5_h * 27) / (float) 1000;
	c_lim->pwr_plus_5_l					=(float) (rep.pwr_plus_5_l * 27) / (float) 1000;
	c_lim->mainboard_plus_1_5_h	=(float) (rep.mainboard_plus_1_5_h * 16) / (float) 1000;
	c_lim->mainboard_plus_1_5_l	=(float) (rep.mainboard_plus_1_5_l * 16) / (float) 1000;
	c_lim->cpu1_vcore_h					=(float) (rep.cpu1_vcore_h * 16) / (float) 1000;
	c_lim->cpu1_vcore_l					=(float) (rep.cpu1_vcore_l * 16) / (float) 1000;
	c_lim->stndby_plus_3_3_h 		=(float) (rep.stndby_plus_3_3_h * 16) / (float) 1000;
	c_lim->stndby_plus_3_3_l 		=(float) (rep.stndby_plus_3_3_l * 16) / (float) 1000;
	c_lim->scsi_plus_2_85_h 		=(float) (rep.scsi_plus_2_85_h * 16) / (float) 1000;
	c_lim->scsi_plus_2_85_l 		=(float) (rep.scsi_plus_2_85_l * 16) / (float) 1000;
	c_lim->sys_temp_h 					=(float) rep.sys_temp_h;
	c_lim->cpu2_temp_h 					=(float) rep.cpu2_temp_h;
	c_lim->cpu1_temp_h 					=(float) rep.cpu1_temp_h;
	c_lim->fan1_speed_l 				=(float) (1350000 / (rep.fan1_speed_l *2));
	c_lim->fan2_speed_l 				=(float) (1350000 / (rep.fan2_speed_l *2));
	c_lim->batt_volt_plus_3_3_h	=(float) (rep.batt_volt_plus_3_3_h * 16) / (float) 1000;
	c_lim->batt_volt_plus_3_3_l	=(float) (rep.batt_volt_plus_3_3_l * 16) / (float) 1000;
	c_lim->stndby_plus_5_h 			=(float) (rep.stndby_plus_5_h * 27) / (float) 1000;
	c_lim->stndby_plus_5_l 			=(float) (rep.stndby_plus_5_l * 27) / (float) 1000;

	return(0);
}

static int	do_get_monitoring_data(CONTEXT *ctx, C_MD *c_md)
{
	unsigned char	req[3];
	unsigned char	rep[20];
	int						rc, len, i;
	M_D						*md;
	unsigned char	cksum = 0;

	memset(c_md, 0, sizeof(C_MD));

	req[0] = 0x53;
	req[1] = CMD_GET_MONITORING_DATA;
	req[2] = req[0] + req[1];
	len = sizeof(rep);
	rc = do_quanta_transaction(ctx, req, 3, rep, &len);
	for (i=0; i < (int)(sizeof(M_D)-1); i++)
		cksum+=rep[i];

	md = (M_D *) rep;

	c_md->cpu2_vcore				=(float) (md->cpu2_vcore *16) / (float) 1000;
	c_md->mainboard_plus_2_7=(float) (md->mainboard_plus_2_7 * 16) / (float) 1000;
	c_md->pwr_plus_3_3			=(float) (md->pwr_plus_3_3 * 16) / (float) 1000;
	c_md->pwr_plus_12				=(float) (md->pwr_plus_12 * 61) / (float) 1000;
	c_md->pwr_minus_12			=(float) (((md->pwr_minus_12 * 16) / ((float)1000 - (float) 2.9016)) / (float) 0.194);
	c_md->pwr_plus_5				=(float) (md->pwr_plus_5 * 27) / (float) 1000;
	c_md->mainboard_plus_1_5=(float) (md->mainboard_plus_1_5 * 16) / (float) 1000;
	c_md->cpu1_vcore				=(float) (md->cpu1_vcore * 16) / (float) 1000;
	c_md->stndby_plus_3_3 	=(float) (md->stndby_plus_3_3 * 16) / (float) 1000;
	c_md->scsi_plus_2_85 		=(float) (md->scsi_plus_2_85 * 16) / (float) 1000;
	c_md->sys_temp 					=(float) md->sys_temp;
	c_md->cpu2_temp 				=(float) md->cpu2_temp;
	c_md->cpu1_temp 				=(float) md->cpu1_temp;
	c_md->fan1_speed 				=(float) (1350000 / (md->fan1_speed *2));
	c_md->fan2_speed 				=(float) (1350000 / (md->fan2_speed *2));
	c_md->batt_volt_plus_3_3=(float) (md->batt_volt_plus_3_3 * 16) / (float) 1000;
	c_md->stndby_plus_5 		=(float) (md->stndby_plus_5 * 27) / (float) 1000;

	return(0);
}

static int	do_reset(CONTEXT *ctx)
{
	unsigned char	req[3];
	unsigned char	rep[16];
	int						rc;
	int						len;

	req[0] = 0x53;
	req[1] = CMD_HARDWARE_RESET;
	req[2] = req[0] + req[1];
	len = sizeof(rep);
	rc = do_quanta_transaction(ctx, req, 3, rep, &len);
	if ((rep[0] == 0) &&
			(rep[1] == 0))
		return(0);
	return(-1);
}

/*
 * Power on the system
 * Returns 0: Success
 *        -1: Failure
 *        -2: Already on
 */
static int	do_power_on(CONTEXT *ctx)
{
	unsigned char	req[3];
	unsigned char	rep[16];
	int						rc;
	int						len;

	req[0] = 0x53;
	req[1] = CMD_POWER_ON;
	req[2] = req[0] + req[1];
	len = sizeof(rep);
	rc = do_quanta_transaction(ctx, req, 3, rep, &len);
	if (rep[0] == 0x01)
		return(-1);
	if (rep[0] == 0x02)
		return(-2);
	return(0);
}

/*
 * Power off the system
 * Returns 0: Success
 *        -1: Failure
 *        -2: Already off
 */
static int	do_power_off(CONTEXT *ctx)
{
	unsigned char	req[3];
	unsigned char	rep[16];
	int						rc;
	int						len;

	req[0] = 0x53;
	req[1] = CMD_POWER_OFF;
	req[2] = req[0] + req[1];
	len = sizeof(rep);
	rc = do_quanta_transaction(ctx, req, 3, rep, &len);
	if (rep[0] == 0x01)
		return(-1);
	if (rep[0] == 0x02)
		return(-2);
	return(0);
}

static int	do_misc_led_on(CONTEXT *ctx)
{
	unsigned char	req[3];
	unsigned char	rep[16];
	int						rc;
	int						len;

	req[0] = 0x53;
	req[1] = CMD_TURN_ON_MISC_LED;
	req[2] = req[0] + req[1];
	len = sizeof(rep);
	rc = do_quanta_transaction(ctx, req, 3, rep, &len);
	if ((rep[0] == 0) &&
			(rep[1] == 0))
		return(0);
	return(-1);
}

static int	do_misc_led_off(CONTEXT *ctx)
{
	unsigned char	req[3];
	unsigned char	rep[16];
	int						rc;
	int						len;

	req[0] = 0x53;
	req[1] = CMD_TURN_OFF_MISC_LED;
	req[2] = req[0] + req[1];
	len = sizeof(rep);
	rc = do_quanta_transaction(ctx, req, 3, rep, &len);
	if ((rep[0] == 0) &&
			(rep[1] == 0))
		return(0);
	return(-1);
}

static int	do_misc_led_toggle(CONTEXT *ctx)
{
	unsigned char	req[3];
	unsigned char	rep[16];
	int						rc;
	int						len;

	req[0] = 0x53;
	req[1] = CMD_TOGGLE_MISC_LED;
	req[2] = req[0] + req[1];
	len = sizeof(rep);
	rc = do_quanta_transaction(ctx, req, 3, rep, &len);
	if ((rep[0] == 0) &&
			(rep[1] == 0))
		return(0);
	return(-1);
}

/*
 * Send a PING to remote hardware.. returns 0 if everything is hunky dory
 */
static int	do_ping(CONTEXT *ctx)
{
	unsigned char	req[3];
	unsigned char	rep[16];
	int						rc;
	int						len;

	req[0] = 0x53;
	req[1] = CMD_PING;
	req[2] = req[0] + req[1];
	len = sizeof(rep);
	rc = do_quanta_transaction(ctx, req, 3, rep, &len);
	if ((rep[0] == 0x53) &&
			(rep[1] == 0x4b) &&
			(rep[2] == 0x9e))
		return(0);
	return(-1);
}

/*
 * Issue a quanta command and get back the response.
 * replen should be set to the expected response data size
 * 
 *
 */
static int	do_quanta_transaction( CONTEXT *ctx,
													unsigned char *req, 
													int						reqlen,
													unsigned char	*rep,
													int						*replen)
{
	int							rc;
	fd_set					read_fds;
	struct timeval	to;
	unsigned char		buffer[4096];
	unsigned char		*p;
	int							len = 0;
	int							waitflag = 1;
#ifdef DEBUG
	unsigned char		*dbg;
	int							dbg_i = 0;
#endif

	memset(rep, 0, *replen);
	memset(buffer, 0, sizeof(buffer));

	flush_buffer(ctx);
	if ((rc = write(ctx->con_sock, req, reqlen))<0)
		{
		perror("write");
		return(-1);
		}
#ifdef DEBUG
printf("Sent\n");
#endif
	p = buffer;
	while(1)
		{
		if (waitflag)
			{
			to.tv_sec = 5;
			to.tv_usec = 0;
			}
		else
			{
			to.tv_sec = 1;
			to.tv_usec = 0;
			}
		FD_ZERO(&read_fds);
		FD_SET(ctx->con_sock, &read_fds);
	
#ifdef DEBUG
		if (waitflag)
			printf("waiting for response\n");
		else
			printf("waiting for beat\n");
#endif

		if ((rc = select((ctx->con_sock+1), &read_fds, NULL, NULL, &to))<0)
			{
			perror("select");
			return(-errno);
			}
		if (!rc)
			break;
		else
			{
			if ((rc = read(ctx->con_sock, p, (sizeof(buffer) - len)))<0)
				{
				perror("read");
				return(-errno);
				}
			if (!rc)
				return(-1);
#if 0
printf("got %d bytes\n",rc);
			for (dbg = p; dbg < (p+rc); dbg++)
				printf("0x%.2x ",*dbg);
printf("\n");
#endif
			len +=rc;
			p+=rc;
			waitflag = 0;
			}
		}
	if (waitflag)
		{
		*replen = 0;
		return(0);
		}
	if (len > *replen)
		{
		memcpy(rep, buffer, *replen);
		*replen = *replen;
		}
	else
		{
		memcpy(rep, buffer, len);
		*replen = len;
		}
#if 0
printf("final\n");
	for (dbg_i = 0; dbg_i < *replen; dbg_i++)
		printf("%.2x ",buffer[dbg_i]);
	printf("\n");
#endif
	return(0);
}

static void flush_buffer(CONTEXT *ctx)
{
	int							rc;
	fd_set					read_fds;
	struct timeval	to;
	unsigned char		buffer[4096];

	to.tv_sec = 0;
	to.tv_usec = 0;
	FD_ZERO(&read_fds);
	FD_SET(ctx->con_sock, &read_fds);
	
	if ((rc = select((ctx->con_sock+1), &read_fds, NULL, NULL, &to))<0)
		{
		perror("select");
		return;
		}
	if (!rc)
		return;
	if ((rc = read(ctx->con_sock, buffer,sizeof(buffer)))<0)
		perror("read");
	return;
}
