
/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <netdb.h>
#ifndef __sgi
#include <linux/types.h>
#include <linux/serial.h>
#endif
#include <libloose.h>
#include <sercon_module_harness.h>
#include <sercon_slave_context.h>
#include <prototypes.h>
#include <sercon_slave_context.h>
#include <pthread.h>

static void	thread_cleanup(void *);
static void	process_msg_from_slave(CONTEXT *);
static int	recv_msg(int , char **, __uint32_t *);
static void	process_sercon_ipc(CONTEXT *, __uint32_t, int, char *, char *);
static int	send_msg(int , char *);
static int	sercon_engage(CONTEXT *);
static void adjust_port_bps(CONTEXT *ctx);
static void	sercon_disengage(CONTEXT *);
static void set_device_address(CONTEXT *ctx, char *port);
static void set_baud(CONTEXT *ctx, unsigned int baud);
static void cc_create_otp(CONTEXT *ctx, CONSOLE_CONNECTION *cc);
static CONSOLE_CONNECTION *cc_create(CONTEXT *ctx);
static int cc_create_socket(CONTEXT *ctx, CONSOLE_CONNECTION *cc);
static void cc_destroy(CONTEXT *ctx, CONSOLE_CONNECTION *cc);
static void cc_cleanup(CONTEXT *ctx);
static CONSOLE_CONNECTION *cc_lookup_by_listen_fd(CONTEXT *ctx, int fd);
static ALERT *alert_create(CONTEXT *ctx);
static void alert_destroy(CONTEXT *ctx, ALERT *alert);
static ALERT *alert_lookup(CONTEXT *ctx, int id);
static int write_alert(CONTEXT *ctx, ALERT *alert);

extern int h_errno;

/*
 * This stuff is so that a thread can 'peek' at another threads context
 * structure (to check for duplicate serial ports, etc)
 */
extern pthread_mutex_t	context_mutex;
extern CONTEXT					*context_head;

static void thread_cleanup(void *data)
{
	CONTEXT *ctx = (CONTEXT *) data;
	
#ifdef THREAD_DEBUG
	printf("Thread %s cleaning up\n",ctx->node_id);
#endif
#ifdef THREAD_DEBUG
	printf("Thread %s disengaging\n",ctx->node_id);
#endif
	sercon_disengage(ctx);
	if (ctx->words)
		fclose(ctx->words);
#ifdef THREAD_DEBUG
	printf("Thread %s disengaged\n",ctx->node_id);
	printf("Thread %s clean\n",ctx->node_id);
#endif
	ctx->refcount--;
}

void *thread_start(void *data)
{
	struct _pthread_cleanup_buffer 	_buff;
	CONTEXT													*ctx = (CONTEXT *) data;
	struct timeval									to;
	fd_set													read_fds;
	int															max, rc;
	char														log_msg[255];
	char														host[128];
	char														domain[128];
	char														address[255];
	struct hostent									*he;

#ifdef THREAD_DEBUG
	sprintf(log_msg,
					"0:0:LOG:Thread %s starting (pid %d)",
					ctx->node_id,
					getpid());
	send_msg(ctx->thread_ipc_write, log_msg);
#endif
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	_pthread_cleanup_push(&_buff, thread_cleanup, (void *) ctx);
	ctx->refcount++;
	if (!(ctx->words = fopen(PKGLIBDIR "/sercon.otp","r")))
		{
		sprintf(log_msg,
						"0:0:LOG: Unable to open %s/sercon.otp for node %s OTP (%m)",
						PKGLIBDIR,
						ctx->node_id);
		send_msg(ctx->thread_ipc_write, log_msg);
		ctx->words= NULL;
		}
	gethostname(host, sizeof(host));
	getdomainname(domain, sizeof(domain));
	if (domain[0] != '\0' && strcasecmp(domain,"(none)"))
		sprintf(address,"%s.%s",host,domain);
	else
		strcpy(address, host);
	
	if (!(he = gethostbyname(address)))
		{
		sprintf(log_msg,
						"0:0:LOG:Unable to lookup '%s'. Connections unavailable (%s)",
						address,
						hstrerror(h_errno));
		send_msg(ctx->thread_ipc_write, log_msg);
		}
	else
		{
		memset(&ctx->module_ipaddr, 0, sizeof(ctx->module_ipaddr));
		memcpy(&ctx->module_ipaddr.s_addr, he->h_addr, sizeof(ctx->module_ipaddr.s_addr));
		}

	while(1)
		{
		CONSOLE_CONNECTION	*cc;

		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(ctx->thread_ipc_read, &read_fds);
		max = ctx->thread_ipc_read;

		if (ctx->con_sock != -1)
			{
			FD_SET(ctx->con_sock, &read_fds);
			if (ctx->con_sock > max)
				max = ctx->con_sock;
			}
		cc = ctx->connections;
		while(cc)
			{
			if (cc->stage == STAGE_LISTENING)
				{
				FD_SET(cc->listen_fd, &read_fds);
				if (cc->listen_fd > max)
					max = cc->listen_fd;
				}
			else if (cc->stage != STAGE_DESTROYED)
				{
				FD_SET(cc->sock_fd, &read_fds);
				if (cc->sock_fd > max)
					max = cc->sock_fd;
				}
			cc = cc -> next;
			}
		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s select() failed (%s)",
							ctx->node_id,
							strerror(errno));
			send_msg(ctx->thread_ipc_write, log_msg);
			pthread_testcancel();
			continue;
			}
		if (!rc)
			{
			pthread_testcancel();
			continue;
			}
		if (FD_ISSET(ctx->thread_ipc_read, &read_fds))
			{
			process_msg_from_slave(ctx);
			}
		if ((ctx->con_sock != -1) && (FD_ISSET(ctx->con_sock, &read_fds)))
			{
			char	cons_bytes[255];
			int		len;

			if ((len = read(ctx->con_sock, cons_bytes, sizeof(cons_bytes)))<0)
				{
				sprintf(log_msg,"0:0:LOG:Unable to read console byte from remote");
				send_msg(ctx->thread_ipc_write, log_msg);
				continue;
				}
			// Do alert matrix stuff here
			{
			ALERT	*alert;
			int		step;

			alert = ctx->alerts;
			while(alert)
				{
				/* We don't want to test for triggers when we're capturing */
				if (alert->capturing)
					{
					alert = alert -> next;
					continue;
					}
				for (step = 0; step < len; step++)
					{
					if (cons_bytes[step] == alert->trigger[alert->current_pos])
						{
						/* Match */
						alert->current_pos++;
						if (alert->current_pos == (int) (strlen(alert->trigger)-1))
							{
							/* Mark for capturing */
							alert->current_pos = 0;
							alert->capturing = 1;
							alert->linecount = 0;
							alert->line_pos = 0;
							alert->cb_pos = 0;
#if 0
							memcpy(&alert->line_buffer[0], &alert->trigger[0],strlen(alert->trigger));
							alert->line_pos = (strlen(alert->trigger)+1);
#endif
							}
						}
					else
						{
						/* No match.. start over with this alert */
						alert->current_pos = 0;
						}
					}
				alert = alert->next;
				}
			// Now actually do the capturing stuff
			alert = ctx->alerts;
			while(alert)
				{
				if (!alert->capturing)
					{
					alert = alert -> next;
					continue;
					}
				for (step = 0; step < len; step++)
					{
					if (cons_bytes[step] != (char) 0x0a)
						{
						if (cons_bytes[step] == (char) 0x0d)
							continue;
						if (!isprint((int) cons_bytes[step]))
							continue;
						/* Just another character */
						alert->line_buffer[alert->line_pos++] = cons_bytes[step];
						}
					else
						{
						/* End of line */
						alert->line_buffer[alert->line_pos] = '\0';
						alert->line_pos = 0;

						/* Copy into our capture buffer */
						alert->capture_buffer[alert->cb_pos] = malloc((strlen(alert->line_buffer)+1));
						strcpy(alert->capture_buffer[alert->cb_pos], alert->line_buffer);
						alert->cb_pos++;
		
						alert->linecount++;
						if (alert->linecount == alert->lines_to_capture)
							{
							int	i;

							write_alert(ctx,alert);
							for (i=0; i < alert->cb_pos; i++)
								free(alert->capture_buffer[i]);
	
							alert->capturing = 0;
							}
						}
					}
				alert = alert->next;
				}
			//  End alert matrix stuff
			}
			cc = ctx->connections;
			while(cc)
				{
				write(cc->sock_fd,cons_bytes, len);
				cc = cc -> next;
				}
			}
		cc = ctx->connections;
		while(cc)
			{
			if (cc->stage == STAGE_LISTENING)
				{
				if (FD_ISSET(cc->listen_fd, &read_fds))
					{
					int	len;

					len = sizeof(struct sockaddr_in);
					if ((cc->sock_fd = accept(cc->listen_fd,
																		(struct sockaddr *)&cc->client_addr,
																		&len))<0)
						{
						sprintf(log_msg,"0:0:LOG:Unable to accept console connection (%m)");
						send_msg(ctx->thread_ipc_write, log_msg);
						cc->stage = STAGE_DESTROYED;
						}
					else
						{
						char	msg[255];
						cc->stage = STAGE_PASSWORD_PROMPT;
						sprintf(msg, "[VACM SERCON] Authentication ->");
						if (write(cc->sock_fd, msg, (strlen(msg)))<0)
							{
							sprintf(log_msg,"0:0:LOG:Unable to write prompt (%m)");
							send_msg(ctx->thread_ipc_write, log_msg);
							cc->stage = STAGE_DESTROYED;
							}
						else
							cc->stage = STAGE_PASSWORD_REPLY;
						} /* connection accepted ok */
					} /* FD set for connection request */
				} /* STAGE_LISTENING */
			else if (cc->stage != STAGE_DESTROYED)
				{
				if (FD_ISSET(cc->sock_fd, &read_fds))
					{
					char	C_byte;
					char	msg[255];
					int		rc;

					if ((rc = read(cc->sock_fd, &C_byte, sizeof(C_byte)))<0)
						{
						sprintf(log_msg,"0:0:LOG:Unable to read msg (%m)");
						send_msg(ctx->thread_ipc_write, log_msg);
						cc->stage = STAGE_DESTROYED;
						}
					else if (rc == 0)
						{
						cc->stage = STAGE_DESTROYED;
						goto next_connection;
						}
					else
						{
						if (cc->stage == STAGE_PASSWORD_REPLY)
							{
							/* Use the conv_buffer to gather bytes */
							if (C_byte != '\n')
								{
								cc->conv_buffer[cc->conv_pos++] = C_byte;
								goto next_connection;
								}
							/* End of line received */
							cc->conv_buffer[(cc->conv_pos-1)] = 0;
							cc->conv_pos = 0;
							if (strcmp(cc->conv_buffer, cc->otpasswd))
								{
								sprintf(log_msg,
												"0:0:LOG:Bad password from %s (us-'%s') (them-'%s')",
												inet_ntoa(cc->client_addr.sin_addr),
												cc->otpasswd,
												cc->conv_buffer);
								send_msg(ctx->thread_ipc_write, log_msg);
								cc->login_retries++;
								if (cc->login_retries == 3)
									{
									sprintf(msg,"[VACM SERCON] Too many retries\n");
									if (write(cc->sock_fd, msg, (strlen(msg)))<0)
										{
										sprintf(log_msg,"0:0:LOG:Unable to write exceeded retries msg (%m)");
										send_msg(ctx->thread_ipc_write, log_msg);
										}
									cc->stage = STAGE_DESTROYED;
									}
								else
									{
									sprintf(msg,"[VACM SERCON] Invalid password\n[VACM SERCON] Authentication->");
									if (write(cc->sock_fd, msg, (strlen(msg)))<0)
										{
										sprintf(log_msg,"0:0:LOG:Unable to write retry prompt (%m)");
										send_msg(ctx->thread_ipc_write, log_msg);
										cc->stage = STAGE_DESTROYED;
										}
									else
										cc->stage = STAGE_PASSWORD_REPLY;
									}
								goto next_connection;
								} /* wrong password */
							/* Good password */
							sprintf(msg,
											"[VACM SERCON] Connecting to console for node '%s' in %s mode\n",
											ctx->node_id,
											(cc->access_mode == 2 ? "Read/Write" : "Read Only"));
							if (write(cc->sock_fd, msg, (strlen(msg)))<0)
								{
								sprintf(log_msg,"0:0:LOG:Unable to write connected msg (%m)");
								send_msg(ctx->thread_ipc_write, log_msg);
								cc->stage = STAGE_DESTROYED;
								}
							else
								{
								cc->stage = STAGE_CONNECTED;
								cc->connect_time = time(NULL);
								}
							goto next_connection;
							}
						else if (cc->stage == STAGE_CONNECTED)
							{
							/* Do byte i/o */
							if (cc->access_mode == MODE_RO)
								goto next_connection; // Skip since he can't write
							if (write(ctx->con_sock, &C_byte, sizeof(C_byte))<0)
								{
								sprintf(msg,"[VACM SERCON] Unable to write to console (%m)\n");
								write(cc->sock_fd, msg, strlen(msg));
								}
							cc->last_byte_time = time(NULL);
							goto next_connection;
							}
						else
							{
							sprintf(log_msg,"0:0:LOG:Data received from connection in bad stage (%d)",cc->stage);
							send_msg(ctx->thread_ipc_write, log_msg);
							cc->stage = STAGE_DESTROYED;
							break;
							} 
						} /* data read OK */
					} /* FD set for conversation */
				} /* STAGE != LISTENING && STAGE != DESTROYED */

next_connection:
			cc = cc -> next;
			}
		/* Cleanup any destroyed connections */
		cc_cleanup(ctx);
		pthread_testcancel();
		}
}

static void process_msg_from_slave(CONTEXT *ctx)
{
	char				*msg_buffer, *cmd, *node_id, *bword, *saved_buffer;
	char				*ipc_cmd, *ipc_args, *variable, *value;
	char				*new_name;
	int					len, rc, client_fd;
	__uint32_t	job_id;
	char				*last = (char *) 0x00;
	char				log_msg[255];

	if ((rc = recv_msg(ctx->thread_ipc_read, &msg_buffer, &len)) < 0)
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s unable to read msg from slave (%s)",
					 	ctx->node_id,
					 	strerror(-rc));
		send_msg(ctx->thread_ipc_write, log_msg);
		pthread_exit(NULL);
		return;
		}
	saved_buffer = malloc(len);
	memcpy(saved_buffer, msg_buffer, len);

#if 0
	printf("Thread %s received msg (%s) from slave\n",
				 ctx->node_id,
				 msg_buffer);
#endif
	if (!(cmd = strtok_r(msg_buffer, ":", &last)))
		goto out;
	if (!(node_id = strtok_r(NULL, ":", &last)))
		goto out;
	if (!strcasecmp(cmd,"NODE_IPC"))
		{
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		job_id = (__uint32_t) strtoul(bword, (char **) NULL, 10);
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(bword);
		if (!(ipc_cmd = strtok_r(NULL, ":", &last)))
			goto out;
		ipc_args = &saved_buffer[((ipc_cmd - msg_buffer)+strlen(ipc_cmd)+1)];
		process_sercon_ipc(ctx, job_id, client_fd, ipc_cmd, ipc_args);
		}
	else if (!strcasecmp(cmd, "NODE_VAR"))
		{
		if (!(variable = strtok_r(NULL, ":", &last)))
			goto out;

		/* BEGIN San mod 07/12/00 */
		value = &saved_buffer[((variable - msg_buffer)+(strlen(variable)+1))];
		/* END San mod 07/12/00 */

		if (!strcasecmp(variable,"DEVICE_ADDRESS"))
			{
			set_device_address(ctx, value);
			sercon_disengage(ctx);
			if (!strcasecmp(value, "NONE"))
				{
				sprintf(log_msg,
								"0:0:LOG:Thread %s disengaged",
								ctx->node_id);
				send_msg(ctx->thread_ipc_write, log_msg);
				}
			else
				{
				if ((rc = sercon_engage(ctx)) < 0)
					{
#ifdef ENGAGE_DEBUG
					sprintf(log_msg,
									"0:0:LOG:Thread %s unable to engaging on %s",
									ctx->node_id,
								ctx->device_address);
					send_msg(ctx->thread_ipc_write, log_msg);
#endif
					}
				}
			}
		else if (!strcasecmp(variable,"BAUDRATE"))
			{
			set_baud(ctx, (unsigned int) atoi(value));
		}
		else if (!strncasecmp(variable,"ALRT",4))
			{
			ALERT	*alert;
			char	*bword, *saved_value, *p;
			int		id,num;

			saved_value = malloc((strlen(value))+1);
			strcpy(saved_value, value);

			last = (char *) 0x00;
			if (!(bword=strtok_r(variable, "-", &last)))
				{
				free(saved_value);
				goto out;
				}
			if (!(bword=strtok_r(NULL, "-", &last)))
				{
				free(saved_value);
				goto out;
				}
			if (!(id = atoi(bword)))
				{
				free(saved_value);
				goto out;
				}
			last = (char *) 0x00;
			if (!(bword=strtok_r(value, "-", &last)))
				{
				free(saved_value);
				goto out;
				}
			if (!(num = atoi(bword)))
				{
				free(saved_value);
				goto out;
				}
			p = &saved_value[(((bword+strlen(bword))-&value[0])+1)];
			alert = alert_create(ctx);
			alert->lines_to_capture = num;
			strncpy(alert->trigger, p, ((sizeof(alert->trigger))-1));

			free(saved_value);
			}
		}
	else if (!strcasecmp(cmd, "CLIENT_DIS"))
		{
		CONSOLE_CONNECTION	*cc;
		char								*fd;
		int									client_fd;
		char								clmsg[255];

		if (!(fd = strtok_r(NULL, ":", &last)))
			goto out;
		client_fd = atoi(fd);

		cc = ctx->connections;
		while(cc)
			{
			if (cc->client_fd == client_fd)
				{
				if ((cc->stage != STAGE_DESTROYED) && (cc->stage != STAGE_LISTENING))
					{
					sprintf(clmsg,"[VACM SERCON] Disconnecting from console because of client disconnection\n");
					write(cc->sock_fd, clmsg, strlen(clmsg));
					}
				cc->stage = STAGE_DESTROYED;
				}
			cc = cc->next;
			}
		cc_cleanup(ctx);
		}
	else if (!strcasecmp(cmd, "NODE_RENAME"))
		{
		char	logfile_old[255];
		char	logfile_new[255];

		if (!(new_name = strtok_r(NULL, ":", &last)))
			goto out;
		/* Rename the logfile */
		sprintf(logfile_old,
						"%s/vacm_sercon_alert_log.%s",
						LOGDIR,
						ctx->node_id);
		free(ctx->node_id);
		ctx->node_id = (char *) malloc(strlen(new_name)+1);
		strcpy(ctx->node_id, new_name);
		sprintf(logfile_new,
						"%s/vacm_sercon_alert_log.%s",
						LOGDIR,
						ctx->node_id);
		if (rename(logfile_old, logfile_new) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s unable to rename logfile (%m)",
							ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			}
		}
	else
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s received unsupported cmd from slave (%s)",
					 	ctx->node_id,
					 	cmd);
		send_msg(ctx->thread_ipc_write, log_msg);
		}

out:
	free(saved_buffer);
	free(msg_buffer);
}

static int write_alert(CONTEXT *ctx, ALERT *alert)
{
	char	filename[255];
	char	log_msg[255];
	FILE	*filep;
	int		i;

	sprintf(filename,
					"%s/vacm_sercon_alert_log.%s",
					LOGDIR,
					ctx->node_id);
	if (!(filep = fopen(filename,"a+")))
		{
		sprintf(log_msg,
						"0:0:LOG:Thread %s Unable to open alert file '%s' (%m)",
						ctx->node_id,
						filename);
		send_msg(ctx->thread_ipc_write, log_msg);
		return(-errno);
		}
	fprintf(filep,"ALERT:%s\n",alert->trigger);
	fprintf(filep,"LINES:%d\n",alert->lines_to_capture);
	for (i=0; i < alert->cb_pos; i++)
		fprintf(filep,"%s\n",alert->capture_buffer[i]);
	fprintf(filep,"ALERT:END\n");
	fclose(filep);
	return(0);
}

static int sercon_engage(CONTEXT *ctx)
{
	char									*last = (char *) 0x00;
	char									log_msg[255];

	if (ctx->con_sock != -1)
		return(-EADDRINUSE);
		
	if (!ctx->device_address)
		return(-EFAULT);

	if (ctx->device_address[0] == '/')
		{
#ifdef ENGAGE_DEBUG
		sprintf(log_msg,
						"0:0:LOG:Thread %s engaging on serial port %s",
						ctx->node_id,
						ctx->device_address);
		send_msg(ctx->thread_ipc_write, log_msg);
#endif
		if ((ctx->con_sock = open(ctx->device_address, O_RDWR)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s engage failed (%m)",
							ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
		adjust_port_bps(ctx);
		}
	else
		{
    struct sockaddr_in  sock;
    int                 I_optval;
    char                address[64];
    unsigned int        port;
    char                *bword;
    char                tmp[64];
    
    strcpy(tmp, ctx->device_address);
    
    if (!(bword = strtok_r(tmp, "-", &last)))
      return(-EFAULT);
    strcpy(address, bword);

    if (!(bword = strtok_r(NULL, "-", &last)))
			return(-EFAULT);

    port = atoi(bword);
#ifdef ENGAGE_DEBUG
		sprintf(log_msg,
						"0:0:LOG:Thread %s engaging on inet addr %s port %d",
						ctx->node_id,
						address,
						port);
		send_msg(ctx->thread_ipc_write, log_msg);
#endif
		if ((ctx->con_sock = socket(AF_INET, SOCK_STREAM, 0))<0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s engage failed (%m)",
							ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
      return(-errno);
			}
    I_optval = 1;
    if (setsockopt(ctx->con_sock,
                   SOL_SOCKET,
                   SO_REUSEADDR,
                   &I_optval,
                   sizeof(I_optval)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s engage failed (%m)",
							ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}

    memset(&sock, 0, sizeof(sock));
    sock.sin_family = AF_INET;
    sock.sin_addr.s_addr = inet_addr(address);
    sock.sin_port = htons(port);
    if (connect(ctx->con_sock,
                (struct sockaddr *) &sock,
                sizeof(sock)) < 0)
			{
			sprintf(log_msg,
							"0:0:LOG:Thread %s engage failed (%m)",
							ctx->node_id);
			send_msg(ctx->thread_ipc_write, log_msg);
			return(-errno);
			}
    }

	return(0);
}

static void adjust_port_bps(CONTEXT *ctx)
{
	struct termios				tty;
	int										baud;
#ifndef __sgi
	struct serial_struct	ser;
#endif

	if (ctx->con_sock == -1)
		return;
  tcgetattr(ctx->con_sock,&tty);
  cfmakeraw(&tty);
	switch((int)ctx->baudrate)
		{
		case 50:
			baud = B50;
			break;
		case 75:
			baud = B75;
			break;
    case 110:
      baud = B110;
      break; 
    case 134:
      baud = B150;
      break; 
    case 200:
      baud = B200;
      break; 
    case 300:
      baud = B300;
      break; 
    case 600:
      baud = B600;
      break; 
    case 1200:
      baud = B1200;
      break; 
    case 1800:
      baud = B1800; 
			break;
    case 2400:
      baud = B2400;
      break; 
    case 4800:
      baud = B4800;
      break; 
    case 9600:
      baud = B9600;
      break; 
    case 19200:
      baud = B19200;
      break; 
    case 38400:
    	baud = B38400;
      break; 
    default: 
      ctx->baudrate=9600;
      baud = B9600;
		}
   cfsetospeed(&tty,baud);
   cfsetispeed(&tty,baud);
   tcsetattr(ctx->con_sock,TCSANOW,&tty);

#ifndef __sgi
	if (ioctl(ctx->con_sock,TIOCGSERIAL,&ser)==0)
		{
    if (!(ser.flags & ASYNC_LOW_LATENCY))
			{
      ser.flags |= ASYNC_LOW_LATENCY;
      ioctl(ctx->con_sock,TIOCSSERIAL,&ser);
      }
    } 
#endif  /*__sgi*/
}

/*
 * Disengage from the serial port
 */
static void sercon_disengage(CONTEXT *ctx)
{
#ifdef ENGAGE_DEBUG
	char	log_msg[255];
	sprintf(log_msg,
					"0:0:LOG:Thread %s disengaging from %s",
						ctx->node_id,
						ctx->device_address);
	send_msg(ctx->thread_ipc_write, log_msg);
#endif
	if (ctx->con_sock != -1)
		{
		int	tmp_sock;
		tmp_sock = ctx->con_sock;
		ctx->con_sock = -1;
		close(tmp_sock);
		}
}

/*
 * Modify the serial port
 */
static void set_device_address(CONTEXT *ctx, char *port)
{
	if (ctx->device_address)
		free(ctx->device_address);
	ctx->device_address = malloc((strlen(port)+1));
	strcpy(ctx->device_address, port);
}

/*
 * Modify the baud rate
 */
static void set_baud(CONTEXT *ctx, unsigned int baud)
{
	ctx->baudrate = baud;
	adjust_port_bps(ctx);
}

/*
 * Process an IPC message
 */
static void process_sercon_ipc(CONTEXT			*ctx,
															__uint32_t	job_id,
															int					client_fd,
															char				*ipc_cmd,
															char				*ipc_args)
{
	char					resp_msg[255];
	char					*last = (char *) 0x00;
	char					log_msg[255];

	if (!strcasecmp(ipc_cmd, "CONFIGURATION"))
		{
		char 					*dev;
		char					*bps;
		unsigned int	baud;
		CONTEXT				*ctx_scan;
		int						dev_changed = 0, bps_changed = 0;

		if (!(dev = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (!(bps = strtok_r(NULL, ":", &last)))
			goto out_malformatted;

		/* Validate the device */
		if (strcasecmp(dev,"NONE"))
			{
			pthread_mutex_lock(&context_mutex);
			ctx_scan = context_head;
			while(ctx_scan)
				{
				if (ctx_scan->device_address)
					if ((!strcmp(ctx_scan->device_address, dev)) &&
							(ctx_scan != ctx))
						break;
				ctx_scan = ctx_scan->next;
				}
			pthread_mutex_unlock(&context_mutex);
			if (ctx_scan)
				{
				sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
							 	strerror(EADDRINUSE));
				send_msg(ctx->thread_ipc_write, resp_msg);
				goto out_ok;
				}
			}

		/* Validate the baud rate */
		baud = atoi(bps);
		switch (baud)
			{
			case 50:
			case 75:
      case 110:
      case 134:
      case 200:
      case 300:
      case 600:
      case 1200:
      case 1800:
      case 2400:
      case 4800:
      case 9600:
      case 19200:
      case 38400:
				break;
      default: 
				sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",
								job_id,
								client_fd,
							 	strerror(EINVAL));
				send_msg(ctx->thread_ipc_write, resp_msg);
				goto out_ok;
			}

		// See if the device or address has changed
		if ((ctx->device_address ? strcmp(ctx->device_address, dev) : 1))
			{
			set_device_address(ctx, dev);
			dev_changed = 1;
			}
		// See if the baudrate has changed
		if (ctx->baudrate != baud)
			{
			set_baud(ctx, baud);
			bps_changed = 1;
			}
		if ((dev_changed) || (bps_changed))
			{
			sprintf(resp_msg,"0:0:SET:%s:SERCON:BAUDRATE:%s",ctx->node_id,bps);
			send_msg(ctx->thread_ipc_write, resp_msg);
			sprintf(resp_msg,"0:0:SET:%s:SERCON:DEVICE_ADDRESS:%s",ctx->node_id,dev);
			send_msg(ctx->thread_ipc_write, resp_msg);

			// Disengage
			sercon_disengage(ctx);
			if (strcasecmp(dev, "NONE"))
				{
				int	rc;
				if ((rc = sercon_engage(ctx) <0))
					{
					sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",
									job_id,
									client_fd,
							 		strerror(EINVAL));
					send_msg(ctx->thread_ipc_write, resp_msg);
					goto out_ok;
					}
				}
			else
				{
				sprintf(log_msg,
								"0:0:LOG:Thread %s disengaging",
								ctx->node_id);
				send_msg(ctx->thread_ipc_write, log_msg);
				}
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	if (ctx->con_sock == -1)
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(ENOPKG));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	if (!strcasecmp(ipc_cmd, "INQUIRY"))
		{
		
		sprintf(resp_msg,
						"%d:%d:FOR:INQUIRY:%s:%s:%d",
						job_id,
						client_fd,
						ctx->node_id,
						ctx->device_address,
						ctx->baudrate);
		send_msg(ctx->thread_ipc_write, resp_msg);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
			
		}
	else if (!strcasecmp(ipc_cmd, "CONNECT")) // Request console connection
		{
		CONSOLE_CONNECTION	*cc;
		char								*mode;
		int									rc;

		if (!(mode = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if ((strcasecmp(mode,"RO")) && (strcasecmp(mode,"RW")))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		cc = cc_create(ctx);
		cc->stage = STAGE_LISTENING;
		cc->client_fd = client_fd;
		if (!strcasecmp(mode,"RW"))
			{
			CONSOLE_CONNECTION	*ccscan;

			ccscan = ctx->connections;
			while(ccscan)
				{
				if (ccscan->access_mode == MODE_RW)
					break;
				ccscan = ccscan->next;
				}
			if (ccscan)
				{
				sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EADDRINUSE));
				send_msg(ctx->thread_ipc_write, resp_msg);
				if (cc->listen_fd != -1)
					close(cc->listen_fd);
				if (cc->sock_fd != -1)
					close(cc->sock_fd);
				cc_destroy(ctx, cc);
				goto out_ok;
				}
			cc->access_mode = MODE_RW;
			}
		else
			cc->access_mode = MODE_RO;
		cc_create_otp(ctx, cc);
		if ((rc = cc_create_socket(ctx, cc))<0)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(-rc));
			send_msg(ctx->thread_ipc_write, resp_msg);
			cc_destroy(ctx, cc);
			goto out_ok;
			}
		sprintf(resp_msg,
						"%d:%d:FOR:CONNECT:%d:%s:%d:%s",
						job_id,
						client_fd,
						cc->listen_fd,
						inet_ntoa(ctx->module_ipaddr),
						cc->port,
						cc->otpasswd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "DISCONNECT")) // Request disconnect
		{
		CONSOLE_CONNECTION	*cc;
		char								*bword;
		char								dis_msg[255];
		int									fd;

		if (!(bword = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (!(fd = atoi(bword)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}

		if (!(cc = cc_lookup_by_listen_fd(ctx, fd)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(ENODEV));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (cc->client_fd != client_fd)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EPERM));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (cc->stage != STAGE_LISTENING)
			{
			sprintf(dis_msg,"[VACM SERCON] Session closed by requesting client\n");
			write(cc->sock_fd, dis_msg, strlen(dis_msg));
			}

		if (cc->listen_fd != -1)
			close(cc->listen_fd);
		if (cc->sock_fd != -1)
			close(cc->sock_fd);
		cc_destroy(ctx, cc);

		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "FORCE_DISCONNECT")) // Force other to disconn
		{
		CONSOLE_CONNECTION	*cc;
		char								*bword;
		char								dis_msg[255];
		int									fd;

		if (!(bword = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (!(fd = atoi(bword)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}

		if (!(cc = cc_lookup_by_listen_fd(ctx, fd)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(ENODEV));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (cc->stage != STAGE_LISTENING)
			{
			sprintf(dis_msg,"[VACM SERCON] Session FORCE closed.\n");
			write(cc->sock_fd, dis_msg, strlen(dis_msg));
			}
		if (cc->listen_fd != -1)
			close(cc->listen_fd);
		if (cc->sock_fd != -1)
			close(cc->sock_fd);
		cc_destroy(ctx, cc);

		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "STEAL_CONNECTION"))
		{
		CONSOLE_CONNECTION	*cc_stealer;
		CONSOLE_CONNECTION	*cc_stealee;
		char								*bword;
		int									fd_stealer;
		int									fd_stealee;
		time_t							current_time;
		char								steal_msg[255];

		if (!(bword = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (!(fd_stealer = atoi(bword)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (!(bword = strtok_r(NULL, ":", &last)))
			goto out_malformatted;
		if (!(fd_stealee = atoi(bword)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (fd_stealee == fd_stealer)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (!(cc_stealer = cc_lookup_by_listen_fd(ctx, fd_stealer)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(ENODEV));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (!(cc_stealee = cc_lookup_by_listen_fd(ctx, fd_stealee)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(ENODEV));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (cc_stealee->access_mode != MODE_RW)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EROFS));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		current_time = time(NULL);
		if ((current_time - cc_stealee->last_byte_time) < (60*5))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EBUSY));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}

		cc_stealee->access_mode = MODE_RO;
		if (cc_stealee->stage != STAGE_LISTENING)
			{
			sprintf(steal_msg,
							"[VACM SERCON] Read/Write mode stolen by client %s\n",
							inet_ntoa(cc_stealer->client_addr.sin_addr));
			write(cc_stealee->sock_fd, steal_msg, strlen(steal_msg));
			}
		if (cc_stealer->stage != STAGE_LISTENING)
			{
			strcpy(steal_msg,"[VACM SERCON] Read/Write mode sucessfully stolen\n");
			write(cc_stealer->sock_fd, steal_msg, strlen(steal_msg));
			}
		cc_stealer->access_mode = MODE_RW;

		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "LIST_CONNECTIONS")) // List current connections
		{
		CONSOLE_CONNECTION	*cc;

		cc = ctx->connections;
		while(cc)
			{
			char			form_time[255];
			struct tm	*tm_time;

			tm_time = localtime(&cc->connect_time);

			strftime(form_time, (sizeof(form_time)-1), "%H %M %S", tm_time);

			sprintf(resp_msg,
							"%d:%d:FOR:CONNECTION:%d:%s:%s:%s",
							job_id,
							client_fd,
							cc->listen_fd,
							inet_ntoa(cc->client_addr.sin_addr),
							form_time,
							(cc->access_mode == MODE_RW ? "RW" : "RO"));
			send_msg(ctx->thread_ipc_write, resp_msg);
			cc = cc->next;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "ALERT_ADD")) // Add ALERT
		{
		ALERT	*alert;
		char	*bword, *p;
		char	*saved_args;
		int		num_lines;
		char	variable_name[16];
		char	variable_value[80];
		char	variable_str[255];

		saved_args = malloc((strlen(ipc_args))+1);
		strcpy(saved_args, ipc_args);

		if (!(bword = strtok_r(ipc_args, ":", &last)))
			{
			free(saved_args);
			goto out_malformatted;
			}
		if (!(num_lines = atoi(bword)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);
			free(saved_args);
			goto out_ok;
			}
		if (num_lines > 1024)
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EOVERFLOW));
			send_msg(ctx->thread_ipc_write, resp_msg);
			free(saved_args);
			goto out_ok;
			}
		p = &saved_args[(((bword+strlen(bword))-&ipc_args[0])+1)];

		alert = ctx->alerts;
		while(alert)
			{
			if (!strcmp(alert->trigger, p))
				break;
			alert = alert -> next;
			}
		if (alert)
			{
			sprintf(resp_msg,"%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EADDRINUSE));
			send_msg(ctx->thread_ipc_write, resp_msg);
			free(saved_args);
			goto out_ok;
			}

		alert = alert_create(ctx);
		alert->lines_to_capture = num_lines;
		strncpy(alert->trigger, p, (sizeof(alert->trigger)-1));
	
		sprintf(variable_name,"ALRT-%d",alert->alert_id);
		sprintf(variable_value,"%d-%s",alert->lines_to_capture,alert->trigger);

		sprintf(variable_str,
						"0:0:SET:%s:SERCON:%s:%s",
						ctx->node_id,
						variable_name,
						variable_value);
		send_msg(ctx->thread_ipc_write, variable_str);

		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		free(saved_args);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "ALERT_DEL")) // Delete ALERT
		{
		ALERT	*alert;
		char	*bword;
		char	variable_str[255];
		char	variable_name[16];
		int		id;

		if (!(bword = strtok_r(ipc_args, ":", &last)))
			goto out_malformatted;
		if (!(id = atoi(bword)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(EINVAL));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		if (!(alert = alert_lookup(ctx, id)))
			{
			sprintf(resp_msg, "%d:%d:FOR:JOB_ERROR:%s",job_id,client_fd,strerror(ENODEV));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}

		sprintf(variable_name,"ALRT-%d",alert->alert_id);
		sprintf(variable_str,
						"0:0:SET:%s:SERCON:%s:ERASE",
						ctx->node_id,
						variable_name);

		send_msg(ctx->thread_ipc_write, variable_str);

		alert_destroy(ctx, alert);
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "ALERT_LIST")) // List ALERTS
		{
		ALERT	*alert;
		char	alert_msg[255];

		alert = ctx->alerts;
		while(alert)
			{
			sprintf(alert_msg,
							"%d:%d:FOR:ALERT:%d:%d:%s",
							job_id,
							client_fd,
							alert->alert_id,
							alert->lines_to_capture,
							alert->trigger);
			send_msg(ctx->thread_ipc_write,alert_msg);
			alert = alert -> next;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "ALERT_DUMP")) // Dump ALERT log
		{
		char	filename[255];
		FILE	*filep;
		char	*bword;
		char	line[1024];
		char	trigger[64];
		int		num, i;
		int		err=0;

		sprintf(filename,
						"%s/vacm_sercon_alert_log.%s",
						LOGDIR,
						ctx->node_id);

		if (!(filep = fopen(filename,"r")))
			{
			if (errno == ENOENT)
				{
				sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
				send_msg(ctx->thread_ipc_write, resp_msg);
				goto out_ok;
				}
			sprintf(resp_msg,
							"%d:%d:FOR:JOB_ERROR:%s",
							job_id,
							client_fd,
							strerror(errno));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		while(1)
			{
			/* Get the trigger */
			if (!fgets(line,sizeof(line),filep))
				{
				err = 0;
				break;
				}
			line[(strlen(line)-1)]=0;
			if (strncmp(line,"ALERT:",6))
				{
				err = 1;
				break;
				}
			bword = &line[6];
			strcpy(trigger,bword);

			/* Get the num lines */
			if (!fgets(line,sizeof(line),filep))
				{
				err = 1;
				break;
				}
			line[(strlen(line)-1)]=0;
			if (strncmp(line,"LINES:",6))
				{
				err = 1;
				break;
				}
			bword = &line[6];
			if (!(num = atoi(bword)))
				{
				err = 1;
				break;
				}
			/* Now get all the lines of data */
			for (i =0; i < num; i++)
				{
				if (!fgets(line,sizeof(line),filep))
					{
					err = 1;
					break;
					}
				line[(strlen(line)-1)]=0;
				sprintf(resp_msg,
								"%d:%d:FOR:ALERT:%s:%d:%d:%s",
								job_id,
								client_fd,
								trigger,
								(i+1),
								num,
								line);
				send_msg(ctx->thread_ipc_write, resp_msg);
				}
			/* Verify this is the termination line */
			if (!fgets(line,sizeof(line),filep))
				{
				err = 1;
				break;
				}
			line[(strlen(line)-1)]=0;
			if (strcmp(line,"ALERT:END"))
				{
				err = 1;
				break;
				}
			}
		fclose(filep);

		if (err)
			{
			sprintf(resp_msg,
							"%d:%d:FOR:JOB_ERROR:%s",
							job_id,
							client_fd,
							strerror(ELIBBAD));
			send_msg(ctx->thread_ipc_write, resp_msg);
			goto out_ok;
			}
		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else if (!strcasecmp(ipc_cmd, "ALERT_CLEAR")) // Clear ALERT log
		{
		char	filename[255];

		sprintf(filename,
						"%s/vacm_sercon_alert_log.%s",
						LOGDIR,
						ctx->node_id);
		unlink(filename);	

		sprintf(resp_msg, "%d:%d:FOR:JOB_COMPLETED",job_id,client_fd);
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}
	else
		{
		sprintf(resp_msg,
						"%d:%d:FOR:JOB_ERROR:%s",
						job_id,
						client_fd,
						strerror(EBADRQC));
		send_msg(ctx->thread_ipc_write, resp_msg);
		goto out_ok;
		}

out_ok:
	return;

out_malformatted:
	sprintf(resp_msg,
					"%d:%d:FOR:JOB_ERROR:%s",
					job_id,
					client_fd,
					strerror(EINVAL));
	send_msg(ctx->thread_ipc_write, resp_msg);
	return;
}

/*
 * Allocate, link in, and assign an alert ID to a new alert
 */
static ALERT *alert_create(CONTEXT *ctx)
{
	ALERT	*new;
	ALERT	*scan;

	new = (ALERT *) malloc(sizeof(ALERT));
	memset(new, 0, sizeof(ALERT));

	if (!ctx->alerts)
		ctx->alerts = new;
	else
		{
		scan = ctx->alerts;
		while(scan->next)
			scan = scan->next;
		scan->next = new;
		new->prev = scan;
		}
	if (ctx->next_alert_id == 0)
		ctx->next_alert_id = 1;
	new->alert_id = ctx->next_alert_id++;
	return(new);
}

/*
 * Destroy and unlink an alert
 */
static void alert_destroy(CONTEXT *ctx, ALERT *alert)
{
	if (alert->prev == NULL)
		{
		if (alert != ctx->alerts)
			{
			printf("SERCON ASSERT FAILURE!!!!\n");
			abort();
			}
		ctx->alerts = alert->next;
		if (ctx->alerts)
			ctx->alerts->prev = NULL;
		}
	else if (alert->next == NULL)
		{
		alert->prev->next = NULL;
		}
	else
		{
		alert->next->prev = alert->prev;
		alert->prev->next = alert->next;
		}
	memset(alert, 0, sizeof(ALERT));
	free(alert);
}

/*
 * lookup an alert by its id
 */
static ALERT *alert_lookup(CONTEXT *ctx, int id)
{
	ALERT	*alert;

	alert = ctx->alerts;
	while(alert)
		{
		if (alert->alert_id == id)
			break;
		alert = alert -> next;
	}
	return(alert);
}


/*
 * Allocate a console connection and link it in
 */ 
static CONSOLE_CONNECTION *cc_create(CONTEXT *ctx)
{
	CONSOLE_CONNECTION	*new;
	CONSOLE_CONNECTION	*scan;

	new = (CONSOLE_CONNECTION *) malloc(sizeof(CONSOLE_CONNECTION));
	memset(new, 0, sizeof(CONSOLE_CONNECTION));

	if (!ctx->connections)
		ctx->connections = new;
	else
		{
		scan = ctx->connections;
		while(scan->next)
			scan = scan->next;
		scan->next = new;
		new->prev = scan;
		}
	new->sock_fd = -1;
	new->listen_fd = -1;
	return(new);
}

static CONSOLE_CONNECTION *cc_lookup_by_listen_fd(CONTEXT *ctx, int fd)
{
	CONSOLE_CONNECTION *cc;

	cc = ctx->connections;
	while(cc)
		{
		if (cc->listen_fd == fd)
			break;
		cc = cc -> next;
	}
	return(cc);
}

static void cc_cleanup(CONTEXT *ctx)
{
	CONSOLE_CONNECTION	*cc;

	cc = ctx->connections;
	while(cc)
		{
		if (cc->stage == STAGE_DESTROYED)
			{
			if (cc->listen_fd != -1)
				close(cc->listen_fd);
			if (cc->sock_fd != -1)
				close(cc->sock_fd);
			cc_destroy(ctx, cc);
			cc = ctx->connections;
			continue;
			}
		cc = cc -> next;
		}
}

/*
 * Create a socket, and bind it to a port
 */
static int cc_create_socket(CONTEXT *ctx, CONSOLE_CONNECTION *cc)
{
	struct sockaddr_in	sock_in;
	char								log_msg[255];
	unsigned short			port_scan = 9000;

  if ((!cc) || (!ctx))
		{
		sprintf(log_msg,"0:0:LOG:cc_create_socket called with bad args");
		send_msg(ctx->thread_ipc_write, log_msg);
		return(-errno);
		}

	if ((cc->listen_fd = socket(AF_INET, SOCK_STREAM, 0))<0)
		{
		sprintf(log_msg,"0:0:LOG:Unable to create connecting socket (%m)");
		send_msg(ctx->thread_ipc_write, log_msg);
		return(-errno);
		}
#if 0
	if (setsockopt(cc->listen_fd,
								 SOL_SOCKET,
								 SO_REUSEADDR,
								 &I_optval,
								 sizeof(I_optval)) < 0)
		{
		sprintf(log_msg,"0:0:LOG:Unable to modify connecting socket (%m)");
		send_msg(ctx->thread_ipc_write, log_msg);
		return(-errno);
		}
#endif
	while(1)
		{
		memset(&sock_in, 0, sizeof(sock_in));
		sock_in.sin_family = AF_INET;
		sock_in.sin_addr.s_addr = htonl(INADDR_ANY);
		sock_in.sin_port = htons(port_scan);
	
		if (bind(cc->listen_fd,
						 (struct sockaddr *) &sock_in,
						 sizeof(sock_in))<0)
			{
			if (errno == EADDRINUSE)
				{
				port_scan++;
				if (port_scan == 10000)
					{
					sprintf(log_msg,"0:0:LOG:Unable to bind connecting socket (%s)",strerror(EUSERS));
					send_msg(ctx->thread_ipc_write, log_msg);
					return(-errno);
					}
				continue;
				}
			sprintf(log_msg,"0:0:LOG:Unable to bind connecting socket (%m)");
			send_msg(ctx->thread_ipc_write, log_msg);
			close(cc->listen_fd);
			return(-errno);
			}
		break;
		}
	cc->port = port_scan;

	if (listen(cc->listen_fd, 5)<0)
		{
		sprintf(log_msg,"0:0:LOG:Unable to listen on connecting socket (%m)");
		send_msg(ctx->thread_ipc_write, log_msg);
		close(cc->listen_fd);
		return(-errno);
		}
	return(0);
}

static void cc_create_otp(CONTEXT *ctx, CONSOLE_CONNECTION *cc)
{
	int						random_fd;
	unsigned char	randomnumber;
	unsigned int	i;
	char					line[255];

	if ((random_fd = open("/dev/random",O_RDONLY))<0)
		randomnumber = (char) random();
	else
		{
		read(random_fd, &randomnumber, 1);
		close(random_fd);
		}
	if (!ctx->words)
		{
		sprintf(cc->otpasswd,"%d",randomnumber);
		return;
		}
	for (i =0; i < randomnumber; i++)
		{
		if (!fgets(line, sizeof(line), ctx->words))
			rewind(ctx->words);
		}
	line[(strlen(line)-1)]=0;
	strncpy(cc->otpasswd, line,(sizeof(cc->otpasswd)-1));
	return;
}

static void cc_destroy(CONTEXT *ctx, CONSOLE_CONNECTION *cc)
{
	if (cc->prev == NULL)
		{
		if (cc != ctx->connections)
			{
			printf("SERCON ASSERT FAILURE!!!!\n");
			abort();
			}
		ctx->connections = cc->next;
		if (ctx->connections)
			ctx->connections->prev = NULL;
		}
	else if (cc->next == NULL)
		{
		cc->prev->next = NULL;
		}
	else
		{
		cc->next->prev = cc->prev;
		cc->prev->next = cc->next;
		}
	memset(cc, 0, sizeof(CONSOLE_CONNECTION));
	free(cc);
}
	
/*
 * Receive a message. This function is threadsafe
 */
static int recv_msg(int fd, char **buffer, __uint32_t *len)
{
  __uint32_t  net_len;
	__uint32_t	brtr;
	char				*offset;
  int         rc;

  if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
	else if (!rc)
		return(-1);
  *len = ntohl(net_len);

  if (!(*buffer = malloc(*len)))
    return(-ENOMEM);
  memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr)
		{
  	if ((rc = read(fd, (void *) offset, brtr)) < 0)
    	return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
		}
  return(0);
}

/*
 * Send a message. This function is threadsafe
 */
static int send_msg(int fd, char *msg)
{
  __uint32_t  len;
  __uint32_t  net_len;
  int         rc;

	if (fd == -1)
		return(-EBADF);
  len = strlen(msg)+1;
  net_len = htonl(len);
  if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
    return(rc);
  if ((rc = write (fd, (void *) msg, len)) < 0)
    return(rc);
  return(0);
}
