/*
 * This program is original work by San Mehat <nettwerk@valinux.com> .
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 1999, San Mehat, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * VACM Remote User Administration Daemon
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <grp.h>
#include <pwd.h>
#include <shadow.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <vacm_user_admd.h>
#include <config.h>

#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#ifndef OPENSSL
#error Due to the nature of this module, OpenSSL encryption is REQUIRED.
#endif

#ifdef HAVE_PAM
#include <security/pam_appl.h>
#include <security/pam_misc.h>

static struct pam_conv	conv;
#endif

#define SSL_HOME "/etc"
#define CERTF SSL_HOME "/vacm.cert"
#define KEYF SSL_HOME "/vacm.key"

#define USERADD_PATH "/usr/sbin/useradd"
#define USERDEL_PATH "/usr/sbin/userdel"
#define GPASSWD_PATH "/usr/bin/gpasswd"
#define USERMOD_PATH "/usr/sbin/usermod"
static int	get_password(void);
static int	set_password(void);
static void	usage(char *name);
static void	process_cmd(NEXXUS *nexxusp, char *buffer);
static void	send_response(__uint32_t job_id, int fd, char *msg);
static SSL	*lookup_ssl_by_fd(int fd);
static SSL	*lookup_ssl_by_fd(int fd);
static int	user_in_group(char *username, char *group);
static void in_progress_handler(int I_signo);
static int	valid_date(char *date);
static void become_daemon(void);

#ifdef HAVE_PAM
static int vacm_verify_conv(int											num_msg,
									   const struct pam_message	**msgm,
									   struct pam_response			**response,
										 void											*appdata_ptr);
#endif

static int				comm_sock;
static NEXXUS			nexxus[255];
static int				debug = 0;
static SSL_CTX		*ctx = NULL;
static SSL_METHOD	*meth = NULL;
static char				daemon_password[255];

int foreground_mode = 0;

int main(argc, argv)
int		argc;
char	**argv;
{
	struct sockaddr_in	sock;
	int									I_optval = 1;
	fd_set							read_fds;
	struct timeval			to;
	int									max, I_c1, len, rc;
	char								buffer[512];
	int									err;

	if (getuid()!=0)
		{
		printf("You must be root to run this\n");
		exit(-1);
		}

	printf("VACM User Administration Daemon version 1.1 (nettwerk@valinux.com)\n");
  while ((rc = getopt(argc, argv, "pf"))!=EOF)
    {
    if ((char)rc == 'p')
      {
      if (!set_password())
        exit(-1);
      exit(0);
      }
    else if ((char)rc =='f')
      foreground_mode = 1;
    else
      {
      usage(argv[0]);
      exit(-1);
      }
    }
  if (!get_password())
    {
    printf("No nexxus access password set. Please set it\n");
    if (!set_password())
			{
			printf("Unable to set password.\n");
      exit(-1);
			}
		printf("Password set. Please rerun daemon\n");
		exit(0);
    } 

	if (!foreground_mode)
		become_daemon();

	signal(SIGPIPE, SIG_IGN);
  SSL_load_error_strings();
  SSLeay_add_ssl_algorithms();
  meth = SSLv23_server_method();
  ctx = SSL_CTX_new(meth);
  if (!ctx)
    {
    ERR_print_errors_fp(stderr);
		exit(-1);
    }
  if (SSL_CTX_use_certificate_file(ctx, CERTF, SSL_FILETYPE_PEM) <=0)
    {
    ERR_print_errors_fp(stderr);
		exit(-1);
    }
  if (SSL_CTX_use_PrivateKey_file(ctx, KEYF, SSL_FILETYPE_PEM) < 0)
    {
    ERR_print_errors_fp(stderr);
		exit(-1);
    }
  if (!SSL_CTX_check_private_key(ctx))
    {
    printf("Private key does not match the certificate public key\n");
		exit(-1);
    }
	if (debug)
		printf("OpenSSL encryption layer enabled\n");

	if ((comm_sock = socket(AF_INET, SOCK_STREAM, 0))<0)
		{
		perror("socket");
		exit(1);
		}
	if (setsockopt(comm_sock,
								 SOL_SOCKET,
								 SO_REUSEADDR,
								 &I_optval,
								 sizeof(I_optval)) < 0)
		{
		perror("setsockopt");
		exit(1);
		}
	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_addr.s_addr = htonl(INADDR_ANY);
	sock.sin_port = htons(NETWORK_PORT);

	if (bind(comm_sock,
					 (struct sockaddr *) &sock,
					 sizeof(sock))<0)
		{
		perror("bind");
		exit(1);
		}
	if (listen(comm_sock, 5)<0)
		{
		perror("listen");
		exit(1);
		}
	for (I_c1=0; I_c1<255;I_c1++)
		nexxus[I_c1].sock = -1;

	while(1)
		{
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;
		FD_SET(comm_sock, &read_fds);
		max = comm_sock;

		for (I_c1 = 0; I_c1 < 255; I_c1++)
			{
			if (nexxus[I_c1].sock != -1)
				{
				if (max < nexxus[I_c1].sock)
					max = nexxus[I_c1].sock;
				FD_SET(nexxus[I_c1].sock, &read_fds);
				}
			}
		if ((rc = select((max+1), &read_fds, NULL, NULL, &to))<0)
			{
			perror("select");
			continue;
			}
		else if (!rc)
			continue;
		else 
			{
			if (FD_ISSET(comm_sock, &read_fds))
				{
				/* New client connection */
				for(I_c1 =0; I_c1<255;I_c1++)
					if (nexxus[I_c1].sock == -1)
						break;
				if (I_c1 == 255)
					{
					printf("[VACM_USER_ADMD] Maximum number of Nexxus's connected\n");
					sleep(1);
					continue;
					}
				len = sizeof(struct sockaddr_in);
				if ((nexxus[I_c1].sock = accept(comm_sock,
																			 (struct sockaddr *) &sock,
																			 &len))<0)
					{
					perror("accept");
					sleep(1);
					continue;
					}
				if (debug)
					{
					printf("[VACM_USER_ADMD] Nexxus connection accepted (fd %d)\n",
								 nexxus[I_c1].sock);
					}
				nexxus[I_c1].ssl = SSL_new(ctx);
				SSL_set_fd(nexxus[I_c1].ssl, nexxus[I_c1].sock);
				err = SSL_accept(nexxus[I_c1].ssl);
    		if (!strcmp(SSL_get_cipher(nexxus[I_c1].ssl), "(NONE)"))
      		{
          printf("Non-SSL client detected. Disconnecting.\n");
      		SSL_free(nexxus[I_c1].ssl);
					nexxus[I_c1].authenticated = 0;
					close(nexxus[I_c1].sock);
					nexxus[I_c1].sock = -1;
					continue;
      		}
				if (debug)
					printf("SSL connection using %s\n",SSL_get_cipher(nexxus[I_c1].ssl));
				if (!(nexxus[I_c1].cert = SSL_get_peer_certificate(nexxus[I_c1].ssl)))
					{
					if (debug)
						printf("Note: Client does not have a certificate\n");
					}
				} /* end new client connection */
			for (I_c1 =0; I_c1<255;I_c1++)
				{
				if (nexxus[I_c1].sock != -1)
					{
					if (FD_ISSET(nexxus[I_c1].sock, &read_fds))
						{
						__uint32_t	len;
		
						if (debug)
							{
							printf("[VACM_USER_ADMD] Data ready on fd %d\n",
										 nexxus[I_c1].sock);
							}
						rc = SSL_read(nexxus[I_c1].ssl, (char *) &len, sizeof(len));
						if ((!rc) || (rc <0))
							{
							SSL_shutdown(nexxus[I_c1].ssl);
							SSL_free(nexxus[I_c1].ssl);
							close(nexxus[I_c1].sock);
							nexxus[I_c1].authenticated = 0;
							nexxus[I_c1].sock = -1;
							continue;
							}
						len = ntohl(len);
						rc = SSL_read(nexxus[I_c1].ssl, (char *) buffer, len);
						if ((!rc) || (rc <0))
							{
							SSL_shutdown(nexxus[I_c1].ssl);
							SSL_free(nexxus[I_c1].ssl);
							close(nexxus[I_c1].sock);
							nexxus[I_c1].authenticated = 0;
							nexxus[I_c1].sock = -1;
							continue;
							}
						process_cmd(&nexxus[I_c1], buffer);
						/* IPC */
						}
					} /* end if valid */
				} /* end for */
			} /* end else */
		} /* end while */
}

static void usage(char *name)
{
	printf("Usage: %s [-p] [-f]\n",name);
  printf("       [-p] flag changes the nexxus access password\n");
  printf("       [-f] flag sets daemon to run in foreground\n");
}

#define GET_NEXT_TOKEN(job_id,dst,fd) \
  { \
  char  *tmp_token; \
  if (!(tmp_token = strtok(NULL,":"))) \
    { \
		send_response(job_id,fd, "ERROR"); \
    return; \
    } \
  strncpy(dst, tmp_token, sizeof(dst)); \
  }

static int	progress_fd = -1;
static char	progress_msg[255];
static __uint32_t progress_id = 0;

static void process_cmd(NEXXUS *nexxusp, char *buffer)
{
	struct passwd	*passwd_entry;
	struct group	*group_entry;
	char					*type;
	char					*cmd;
	char					*id;
	int						i;
	int						fd;
	__uint32_t		job_id;

	fd = nexxusp->sock;

	if (debug)
		printf("[VACM_USER_ADMD] Processing <%s>\n",buffer);
	if (!(type = strtok(buffer,":")))
		return;
	if (!nexxusp->authenticated)
		{
		if (!strncmp(type,"AUTH",4))
			{
			char	*password;
			char	*crypted;

			if (!(password = strtok(NULL,":")))
				goto bad_pass;

			crypted = crypt(password, "HI");
			if (strcmp(crypted, daemon_password))
				goto bad_pass;
			nexxusp->authenticated = 1;
			send_response(0, fd, "AUTH_OK");
			return;
bad_pass:
			send_response(0, fd, "AUTH_ERROR");
			return;
			}
		else
			{
			printf("Received msg type <%s> when not authenticated! Naughty...\n",type);
			return;
			}
		}
	if (!(cmd = strtok(NULL,":")))
		return;
	if (strncmp(type,"REQUEST", 7))
		return;

	if (*(cmd + strlen(cmd) + 1) == (char)NULL)
		return;
	if ( (id = strrchr( (cmd + strlen(cmd) + 1), ':')) == NULL)
		id = cmd + strlen(cmd) + 1;
	else
		id++;
	
	job_id = (__uint32_t) strtoul(id, (char **) NULL, 10);
	if (debug)
		printf("job_id: %d\n", job_id);

	if (!strncmp(cmd,"USER_ADD",8))
		{
    char  user_name[16];
    char  user_id[16];
    char  primary_group[32];
    char  home_dir[255];
    char  shell_path[255];
    char  password[255];
    char  expire_date[16];
    char  inactive_days[8];
    char  comment[42];
    char	command_line[512];
		char	tmp[255];

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,user_id,fd);
    GET_NEXT_TOKEN(job_id,primary_group,fd);
    GET_NEXT_TOKEN(job_id,home_dir,fd);
    GET_NEXT_TOKEN(job_id,shell_path,fd);
    GET_NEXT_TOKEN(job_id,password,fd);
    GET_NEXT_TOKEN(job_id,expire_date,fd);
    GET_NEXT_TOKEN(job_id,inactive_days,fd);
    GET_NEXT_TOKEN(job_id,comment,fd);

		snprintf(command_line,sizeof(command_line),"%s ",USERADD_PATH);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if ((passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_IN_USE");
			goto send_done_err;
			}

		if (strcmp(comment,"NONE"))
			{
			/* Verify the comment */
			for (i=0; i < (int) strlen(comment); i++)
				if ((!isprint((int)comment[i])) || (comment[i] == ':'))
					{
					send_response(job_id,fd,"JOB_ERROR:BAD_COMMENT");
					goto send_done_err;
					}
			if ((int) strlen(comment) > 40)
				{
				send_response(job_id,fd,"JOB_ERROR:COMMENT_TOO_LONG");
				goto send_done_err;
				}
			snprintf(tmp,sizeof(tmp),"-c \"%s\" ",comment);
			strcat(command_line, tmp);
			}
		if (strcmp(home_dir,"DEFAULT"))
			{
			for (i=0; i < (int) strlen(home_dir); i++)
				if (!isprint((int) home_dir[i]))
					{
					send_response(job_id,fd,"JOB_ERROR:BAD_HDIR_PATH");
					goto send_done_err;
					}
			snprintf(tmp,sizeof(tmp),"-d %s ",home_dir);
			strcat(command_line, tmp);
			}
		if (strcmp(expire_date,"NONE"))
			{
			if (!valid_date(expire_date))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_DATE");
				goto send_done_err;
				}
			snprintf(tmp,sizeof(tmp),"-e %s ",expire_date);
			strcat(command_line, tmp);
			}
		if (strcmp(inactive_days,"NONE"))
			{
			snprintf(tmp,sizeof(tmp),"-f %s ",inactive_days);
			strcat(command_line, tmp);
			}
		if (strcmp(primary_group,"DEFAULT"))
			{
			if (!atoi(primary_group))
				{
				if ((primary_group[0] == '0') && (primary_group[1] == (char) 0))
					{
					/* The number 0 */
					if (!(group_entry = getgrgid(atoi(primary_group))))
						{
						send_response(job_id,fd,"JOB_ERROR:GRP_NOT_FOUND");
						goto send_done_err;
						}
					}
				else
					{
					/* A string name */
					if (!(group_entry = getgrnam(primary_group)))
						{
						send_response(job_id,fd,"JOB_ERROR:GRP_NOT_FOUND");
						goto send_done_err;
						}
					}
				}
			else
				{
				/* A GID */
				if (!(group_entry = getgrgid(atoi(primary_group))))
					{
					send_response(job_id,fd,"JOB_ERROR:GRP_NOT_FOUND");
					goto send_done_err;
					}
				}
			snprintf(tmp,sizeof(tmp),"-g %s ",primary_group);
			strcat(command_line, tmp);
			}
		if (strcmp(password,"NONE"))
			{
			snprintf(tmp,sizeof(tmp),"-p \"%s\" ",password);
			strcat(command_line, tmp);
			}
		if (strcmp(shell_path,"DEFAULT"))
			{
			snprintf(tmp,sizeof(tmp),"-s %s ",shell_path);
			strcat(command_line, tmp);
			}
		if (strcmp(user_id,"DEFAULT"))
			{
			/* Verify the UID */
			if ((!atoi(user_id)) && (user_id[0] !='0'))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_UID");
				goto send_done_err;
				}
			if ((passwd_entry = getpwuid(atoi(user_id))))
				{
				send_response(job_id,fd,"JOB_ERROR:UID_IN_USE");
				goto send_done_err;
				}
			snprintf(tmp,sizeof(tmp),"-u %s ",user_id);
			strcat(command_line, tmp);
			}
		strcat(command_line,user_name);
#if 0
		printf("Command line is <%s>\n",command_line);
#endif
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_REMOVE", 11))
		{
    char  user_name[16];
    char  delete_flag[16];
		int		pid, status;

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,delete_flag,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry=getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		if (!strcasecmp(user_name,"root"))
			{
			send_response(job_id,fd,"JOB_ERROR:PERMISSION_DENIED");
			goto send_done_err;
			}
		if ((strcasecmp(delete_flag,"YES"))&&(strcasecmp(delete_flag,"NO")))
			{
			send_response(job_id,fd,"JOB_ERROR:MALFORMATTED_MESSAGE");
			goto send_done_err;
			}
		signal(SIGALRM, in_progress_handler);
		strcpy(progress_msg,"USER_REMOVE");
		progress_id = job_id;
		progress_fd = fd;
		if (!(pid = fork()))
			{
			if (!strcasecmp(delete_flag,"YES"))
				execl(USERDEL_PATH, USERDEL_PATH, "-r", user_name, NULL);
			else
				execl(USERDEL_PATH, USERDEL_PATH, user_name, NULL);
			}
		alarm(2);
		waitpid(pid, &status, 0);
		alarm(0);
		progress_fd = -1;
		signal(SIGALRM, SIG_DFL);
		if (!WIFEXITED(status))
			{
			send_response(job_id,fd,"USER_REMOVE:PROCESS_FAILURE");
			send_response(job_id,fd, "ERROR"); 
			return;
			}
		}
	else if (!strncmp(cmd, "USER_GRP_ADD", 12))
		{
    char  user_name[16];
    char  group_name[16];
    char	command_line[512];
		int		rc;

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,group_name,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		if (!(group_entry = getgrnam(group_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:GRP_NOT_FOUND");
			goto send_done_err;
			}
		if ((rc = user_in_group(user_name, group_name))<0)
			{
			send_response(job_id,fd, "ERROR"); 
			return;
			}
		if (rc)
			{
			send_response(job_id,fd,"JOB_ERROR:ALREADY_MEMBER");
			goto send_done_err;
			}
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -a %s %s",
						 GPASSWD_PATH,
						 user_name,
						 group_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_GRP_REM", 12))
		{
    char  user_name[16];
    char  group_name[16];
    char	command_line[512];
		int		rc;

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,group_name,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		if (!(group_entry = getgrnam(group_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:GRP_NOT_FOUND");
			goto send_done_err;
			}
		if ((rc = user_in_group(user_name, group_name))<0)
			{
			send_response(job_id,fd, "ERROR"); 
			return;
			}
		if (rc == 2)
			{
			send_response(job_id,fd,"JOB_ERROR:PRIMARY_GROUP");
			goto send_done_err;
			}
		
		if (!rc)
			{
			send_response(job_id,fd,"JOB_ERROR:NOT_MEMBER");
			goto send_done_err;
			}
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -d %s %s",
						 GPASSWD_PATH,
						 user_name,
						 group_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_CHG_GRP", 12))
		{
    char 	user_name[16];
    char 	group_name[16];
    char	command_line[512];
		int		rc;

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,group_name,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		if (!(group_entry = getgrnam(group_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:GRP_NOT_FOUND");
			goto send_done_err;
			}
		if ((rc = user_in_group(user_name, group_name))<0)
			{
			send_response(job_id,fd, "ERROR"); 
			return;
			}
		if (rc == 2)
			{
			send_response(job_id,fd,"JOB_ERROR:ALREADY_PRIMARY");
			goto send_done_err;
			}
		if (rc)
			{
			/* Remove the user from this group */
			snprintf(command_line,
							 sizeof(command_line),
							 "%s -d %s %s",
							 GPASSWD_PATH,
							 user_name,
							 group_name);
			system(command_line);
			}
		/* Set the primary gid */
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -g %s %s",
						 USERMOD_PATH,
						 group_name,
						 user_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_CHG_DIR", 12))
		{
    char 	user_name[16];
    char 	home_dir[255];
    char	command_line[512];

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,home_dir,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(home_dir); i++)
			if (!isprint((int) home_dir[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_HDIR_PATH");
				goto send_done_err;
				}
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -d %s %s",
						 USERMOD_PATH,
						 home_dir,
						 user_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_CHG_SHELL", 14))
		{
    char 	user_name[16];
    char 	shell_path[255];
    char	command_line[512];

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,shell_path,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(shell_path); i++)
			if (!isprint((int) shell_path[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_SHELL_PATH");
				goto send_done_err;
				}
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -s %s %s",
						 USERMOD_PATH,
						 shell_path,
						 user_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_CHG_UID", 12))
		{
    char 	user_name[16];
    char 	user_id[16];
		int		pid, status;

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,user_id,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		if ((!atoi(user_id)) && (user_id[0] !='0'))
			{
			send_response(job_id,fd,"JOB_ERROR:BAD_UID");
			goto send_done_err;
			}
		if (atoi(user_id) == 0)
			{
			send_response(job_id,fd,"JOB_ERROR:PERMISSION_DENIED");
			goto send_done_err;
			}
		if ((passwd_entry = getpwuid(atoi(user_id))))
			{
			send_response(job_id,fd,"JOB_ERROR:UID_IN_USE");
			goto send_done_err;
			}
		signal(SIGALRM, in_progress_handler);
		strcpy(progress_msg,"USER_CHG_UID");
		progress_id = job_id;
		progress_fd = fd;
		if (!(pid = fork()))
			execl(USERMOD_PATH, USERMOD_PATH, "-u", user_id, user_name, NULL);
		alarm(2);
		waitpid(pid, &status, 0);
		alarm(0);
		progress_fd = -1;
		signal(SIGALRM, SIG_DFL);
		if (!WIFEXITED(status))
			{
			send_response(job_id,fd,"USER_CHG_UID:PROCESS_FAILURE");
			send_response(job_id,fd, "ERROR"); 
			return;
			}
		}
	else if (!strncmp(cmd, "USER_CHG_EXP", 12))
		{
    char 	user_name[16];
    char  expire_date[16];
    char	command_line[512];

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,expire_date,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(expire_date); i++)
			if (!isprint((int) expire_date[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_DATE");
				goto send_done_err;
				}
		if (!valid_date(expire_date))
			{
			send_response(job_id,fd,"JOB_ERROR:BAD_DATE");
			goto send_done_err;
			}
		snprintf(command_line,
					   sizeof(command_line),
						 "%s -e %s %s",
						 USERMOD_PATH,
						 expire_date,
						 user_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_CHG_INA", 12))
		{
    char 	user_name[16];
    char  inactive_days[8];
    char	command_line[512];

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,inactive_days,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(inactive_days); i++)
			if (!isprint((int) inactive_days[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_DATE");
				goto send_done_err;
				}
		if ((!atoi(inactive_days)) && (inactive_days[0]!='0'))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_VALUE");
				goto send_done_err;
				}
			
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -f %s %s",
						 USERMOD_PATH,
						 inactive_days,
						 user_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_CHG_COMMENT", 16))
		{
    char 	user_name[16];
    char  comment[42];
    char	command_line[512];

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,comment,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(comment); i++)
			if ((!isprint((int)comment[i])) || (comment[i] == ':'))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_COMMENT");
				goto send_done_err;
				}
		if ((int) strlen(comment) > 40)
			{
			send_response(job_id,fd,"JOB_ERROR:COMMENT_TOO_LONG");
			goto send_done_err;
			}
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -c \"%s\" %s",
						 USERMOD_PATH,
						 comment,
						 user_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_CHG_PASSWD", 15))
		{
    char 					user_name[16];
    char  				decrypted_oldpasswd[64];
    char  				decrypted_newpasswd[64];
#ifdef HAVE_PAM
		int						rc;
		pam_handle_t	*pamh=NULL;
#endif

    GET_NEXT_TOKEN(job_id,user_name,fd);
    GET_NEXT_TOKEN(job_id,decrypted_oldpasswd,fd);
    GET_NEXT_TOKEN(job_id,decrypted_newpasswd,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(decrypted_oldpasswd); i++)
			if ((!isprint((int)decrypted_oldpasswd[i])) || (decrypted_oldpasswd[i] == ':'))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_PASSWORD");
				goto send_done_err;
				}
		for (i=0; i < (int) strlen(decrypted_newpasswd); i++)
			if ((!isprint((int)decrypted_newpasswd[i])) || (decrypted_newpasswd[i] == ':'))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_PASSWORD");
				goto send_done_err;
				}
#ifdef HAVE_PAM
		conv.conv = vacm_verify_conv;
		conv.appdata_ptr = strdup(decrypted_oldpasswd);
		if ((rc = pam_start("vacm",
												user_name,
												&conv,
												&pamh))!=PAM_SUCCESS)
			{
			send_response(job_id,fd,"ERROR");
			return;
			}
		if ((rc = pam_authenticate(pamh, 0))!=PAM_SUCCESS)
			{
			pam_end(pamh, PAM_SUCCESS);
			send_response(job_id,fd,"JOB_ERROR:PERMISSION_DENIED");
			goto send_done_err;
			}
		pam_end(pamh, PAM_SUCCESS);
		/* Now modify it */
		conv.conv = vacm_verify_conv;
		conv.appdata_ptr = strdup(decrypted_newpasswd);
		if ((rc = pam_start("vacm",
												user_name,
												&conv,
												&pamh))!=PAM_SUCCESS)
			{
			send_response(job_id,fd,"ERROR");
			return;
			}
		if ((rc = pam_chauthtok(pamh, 0)) != PAM_SUCCESS)
			{
			char	err_str[255];

			snprintf(err_str,sizeof(err_str),"JOB_ERROR:PAM_ERR:%s",pam_strerror(pamh,rc));
			send_response(job_id,fd,err_str);
			pam_end(pamh, PAM_SUCCESS);
			goto send_done_err;
			}
		pam_end(pamh, PAM_SUCCESS);
#else
		send_response(job_id,fd,"JOB_ERROR:NO_PAM_SUPPORT");
		goto send_done_err;
#endif
		}
	else if (!strncmp(cmd, "USER_LIST", 9))
		{
		struct spwd	*shadow_entry;
		char				reply[512];
		char				expiry[64];
		char				inactive[64];
		char				group_name[64];
		time_t			time_expire;
		struct tm		*time_val;

		setpwent();
		while((passwd_entry = getpwent()))
			{
			strcpy(expiry,"UNKNOWN");
			strcpy(inactive,"UNKNOWN");
			strcpy(group_name,"UNKNOWN");
			if ((shadow_entry=getspnam(passwd_entry->pw_name)))
				{
				if (shadow_entry->sp_expire == -1)
					strcpy(expiry,"NONE");
				else
					{
					time_expire = (shadow_entry->sp_expire * (24*60*60));
					time_val =localtime(&time_expire);
					strftime(expiry,sizeof(expiry),"%Y-%m-%d",time_val);
					}
				if (shadow_entry->sp_inact == -1)
					strcpy(inactive,"NONE");
				else
					snprintf(inactive,sizeof(inactive),"%d",(int) shadow_entry->sp_inact);
				}
			if ((group_entry=getgrgid(passwd_entry->pw_gid)))
				strcpy(group_name,group_entry->gr_name);
			
			snprintf(reply,
							 sizeof(reply),
							"USER_LIST:%s:%d:%s/%d:%s:%s:%s:%s:%s",
							passwd_entry->pw_name,
							passwd_entry->pw_uid,
							group_name,
							passwd_entry->pw_gid,
							(passwd_entry->pw_dir[0] ? passwd_entry->pw_dir : "NONE"),
							(passwd_entry->pw_shell[0] ? passwd_entry->pw_shell : "NONE"),
							expiry,
							inactive,
							passwd_entry->pw_gecos);
			send_response(job_id,fd, reply);
			}
		endpwent();
		}
	else if (!strncmp(cmd, "GROUP_LIST", 10))
		{
		char	reply[512];

		setgrent();
		while((group_entry = getgrent()))
			{
			snprintf(reply,
							 sizeof(reply),
							 "%s:%d",
							 group_entry->gr_name,
							 (int) group_entry->gr_gid);
			send_response(job_id,fd, reply);
			}
		endgrent();
		}
	else if (!strncmp(cmd, "LOCK_USER", 9))
		{
    char 	user_name[16];
    char	command_line[512];

    GET_NEXT_TOKEN(job_id,user_name,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -L %s",
						 USERMOD_PATH,
						 user_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "UNLOCK_USER", 11))
		{
    char 	user_name[16];
    char	command_line[512];

    GET_NEXT_TOKEN(job_id,user_name,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		snprintf(command_line,
						 sizeof(command_line),
						 "%s -U %s",
						 USERMOD_PATH,
						 user_name);
		system(command_line);
		}
	else if (!strncmp(cmd, "USER_GRP_LIST", 13))
		{
    char 	user_name[16];
		char	reply[255];
		int		i;

    GET_NEXT_TOKEN(job_id,user_name,fd);

		/* Verify the user name */
		if ((int) strlen(user_name) > 8)
			{
			send_response(job_id,fd,"JOB_ERROR:NAME_TOO_LONG");
			goto send_done_err;
			}
		for (i=0; i < (int) strlen(user_name); i++)
			if (!isalpha((int) user_name[i]))
				{
				send_response(job_id,fd,"JOB_ERROR:BAD_NAME");
				goto send_done_err;
				}
		if (!(passwd_entry = getpwnam(user_name)))
			{
			send_response(job_id,fd,"JOB_ERROR:USER_NOT_FOUND");
			goto send_done_err;
			}
		/* First send up the primary group */
		group_entry = getgrgid(passwd_entry->pw_gid);
		snprintf(reply,
						 sizeof(reply),
						 "USER_GRP_LIST:%s:%d",
						 group_entry->gr_name,
						 group_entry->gr_gid);
		send_response(job_id,fd, reply);
		/* Then any secondary groups */
		setgrent();
		while((group_entry = getgrent()))
			{
			for (i=0;;i++)
				{
				if (!group_entry->gr_mem[i])
					break;
				if (!strncmp(group_entry->gr_mem[i], user_name, strlen(user_name)))
					{
					snprintf(reply,
									 sizeof(reply),
									 "USER_GRP_LIST:%s:%d",
									 group_entry->gr_name,
									 group_entry->gr_gid);
					send_response(job_id,fd, reply);
					break;
					}
				}
			}
		endgrent();
		}

	send_response(job_id,fd, "DONE");
	return;
send_done_err:
	send_response(job_id,fd, "DONE_ERR");
	return;
}

#ifdef HAVE_PAM
static int vacm_verify_conv(int												num_msg,
									   				const struct pam_message	**msg,
									   				struct pam_response				**response,
										 				void											*appdata_ptr)
{
	struct pam_response *reply = NULL;
	int									replies = 0;
	

	if (!(reply = malloc(sizeof(struct pam_response) * num_msg)))
		return(PAM_CONV_ERR);
	for (replies = 0; replies < num_msg; replies++)
		{
		switch (msg[replies]->msg_style)
			{
			case PAM_PROMPT_ECHO_ON:
				return(PAM_CONV_ERR);
				break;
			case PAM_PROMPT_ECHO_OFF:
#if 0
				fprintf(stdout,"echo off msg = <%s>\n",msg[replies]->msg);
#endif
				reply[replies].resp_retcode = PAM_SUCCESS;
				reply[replies].resp = strdup(appdata_ptr);
				break;
			case PAM_TEXT_INFO:
				reply[replies].resp_retcode = PAM_SUCCESS;
				reply[replies].resp = NULL;
				break;
			case PAM_ERROR_MSG:
				reply[replies].resp_retcode = PAM_SUCCESS;
				reply[replies].resp = NULL;
				break;
			default:
				return(PAM_CONV_ERR);
			}
		}
	*response = reply;
	return(PAM_SUCCESS);
}
#endif

/* YYYY-MM-DD */
static int valid_date(char *date)
{
	char	date_copy[16];
	char	*bword;
	int 	i;

	memcpy(&date_copy[0], date, strlen(date));

	if (!(bword = strtok(date_copy, "-")))
		goto bad_date;
	if ((int)strlen(bword) != 4)
		goto bad_date;
	for (i=0; i < 4; i++)
		if (!isdigit((int) bword[i]))
			goto bad_date;

	if (!(bword = strtok(NULL, "-")))
		goto bad_date;
	if ((int)strlen(bword) != 2)
		goto bad_date;
	for (i=0; i <= 1; i++)
		if (!isdigit((int) bword[i]))
			goto bad_date;
	if ((!atoi(bword)) || (atoi(bword) > 12))
		goto bad_date;
	
	if (!(bword = strtok(NULL, "-")))
		goto bad_date;
	if ((int)strlen(bword) != 2)
		goto bad_date;
	for (i=0; i <= 1; i++)
		if (!isdigit((int) bword[i]))
			goto bad_date;
	if ((!atoi(bword)) || (atoi(bword) > 31))
		goto bad_date;
	return(1);
bad_date:
	return(0);
}

static void in_progress_handler(int I_signo)
{
  if (I_signo == SIGALRM)
    {
	  char	msg[255];

	  if (progress_fd == -1)
	  	return;
	  snprintf(msg,sizeof(msg),"%s:IN_PROGRESS",progress_msg);
	  send_response(progress_id,progress_fd, msg);
	  alarm(2);
    }
}

static void send_response(__uint32_t job_id,int fd, char *msg)
{
	__uint32_t	len;
	__uint32_t	netlen;
	char				finalmsg[255];
	int					rc;
	SSL					*ssl;

	if (debug)
		printf("[VACM_USER_ADMD] Sending response <%s>\n",msg);
	if (job_id == 0)
		strcpy(finalmsg, msg);
	else
		sprintf(finalmsg,"%d:%s",job_id,msg);
		
	len = strlen(finalmsg)+1;
	netlen = htonl(len);

	ssl = lookup_ssl_by_fd(fd);
  rc = SSL_write(ssl, (char *) &netlen, sizeof(netlen));
	if ((!rc) || (rc <0))
		{
		ERR_print_errors_fp(stderr);
		return;
		}
  rc = SSL_write(ssl, finalmsg, (int) len);
	if ((!rc) || (rc <0))
		{
		ERR_print_errors_fp(stderr);
		return;
		}
}

static SSL *lookup_ssl_by_fd(int fd)
{
	int	I_c1;

	for (I_c1=0; I_c1<255;I_c1++)
		if (nexxus[I_c1].sock == fd)
			return(nexxus[I_c1].ssl);
  return(NULL);
}

static int user_in_group(char *username, char *group)
{
	struct group	*group_entry;
	struct passwd	*passwd_entry;
	int						i = 0;

	if (!(group_entry = getgrnam(group)))
		return(-1);
	if (!(passwd_entry = getpwnam(username)))
		return(-1);

	if (group_entry->gr_gid == passwd_entry->pw_gid)
		return(2); /* primary member */

	while(1)
		{
		if (!group_entry->gr_mem[i])
			break;
		if (!strncmp(group_entry->gr_mem[i], username, strlen(username)))
			return(1);
		i++;
		}
	return(0);
}

static int get_password()
{
  FILE  *Tfilep_ptr; 
    
  if (!(Tfilep_ptr = fopen("/etc/vacm_user_admd.passwd","r")))
    {
    printf("Unable to open /etc/vacm_user_admd.passwd (%m)\n");
    return(0);
    }
  if (!fgets(daemon_password, sizeof(daemon_password), Tfilep_ptr))
    {
    fclose(Tfilep_ptr);
    return(0);
    }
  daemon_password[(strlen(daemon_password)-1)]=0;
  fclose(Tfilep_ptr);
  return(1);
}

static int set_password()
{
  FILE            *Tfilep_ptr;
  struct termios  term;
  char            passwd1[255];
  char            passwd2[255];
  char            *crypted;
  
  unlink("/etc/vacm_user_admd.passwd");
  if (!(Tfilep_ptr = fopen("/etc/vacm_user_admd.passwd","w")))
    {
    printf("Unable to create /etc/vacm_user_admd.passwd (%m)\n");
    return(0);
    }
    
  tcgetattr(fileno(stdin), &term);
  term.c_lflag &= ~ECHO;
  tcsetattr(fileno(stdin), TCSANOW, &term);
  
  fprintf(stdout,"Enter new password -=>");
  fflush(stdout);
  fgets(passwd1,sizeof(passwd1),stdin);
  passwd1[(strlen(passwd1)-1)]=0;
  fprintf(stdout,"\nRe-enter new password -=>");
  fgets(passwd2,sizeof(passwd2),stdin);
  passwd2[(strlen(passwd2)-1)]=0;
  term.c_lflag |= ECHO;
  tcsetattr(fileno(stdin), TCSANOW, &term);
  printf("\n");
  if (strcmp(passwd1, passwd2))
    {
    printf("Passwords do not match\n");
    fclose(Tfilep_ptr);
    return(0);
    }
  crypted = crypt(passwd1, "HI");
  memset(passwd1,0,sizeof(passwd1));
  memset(passwd2,0,sizeof(passwd2));
  fprintf(Tfilep_ptr,"%s\n",crypted);
  strcpy(daemon_password, crypted);
  fclose(Tfilep_ptr);
  return(1);
}

static void become_daemon()
{ 
  int i;
  int fd;
  
  chdir("/");
  if (fork())
    _exit(0);
  setsid(); 
  close(0); 
  close(1); 
  close(2); 
  for (i=0; i < 3; i++)
    {
    fd = open("/dev/null", O_RDWR, 0);
    if (fd < 0) 
      fd = open("/dev/null", O_WRONLY,0);
    if (fd < 0)
      {
      fprintf(stderr,"Unable to detach\n");
      _exit(-1);
      }
   if (fd != i)
      {
      fprintf(stderr,"Unable to detach\n");
      _exit(-1);
      }
    } 
}   
 
