/*
 * This program was original work by San Mehat <nettwerk@valinux.com> 
 * It has since been largely rewritten and is currently maintained by 
 * Zac Sprackett <zsprackett@valinux.com>.
 *
 * This program is distributed under the GNU General Public License (GPL)
 * as outlined in the COPYING file.
 *
 * Copyright (C) 2000, San Mehat, Zac Sprackett, and VA Linux Systems
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Notes: This is now a stand-alone run-time linkable module for VACM
 * Symbols beginning with 'api' denote a module harness required
 * symbol which should not be renamed. Renaming an 'api' symbol
 * will cause undesired operation, and program failure.
 *
 */

#include <config.h>

#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/stat.h>
#include <netdb.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <vacmclient_api.h>
#ifdef HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include <vash.h>

#define INPUT_BLOCK_SIZE 512

char          *prompt = "vash$ ";
char					*inputbuff;
int						inputbuffsize;
int						interactive = 1;
int						job_counter = 0;
struct handle *first_handle = NULL;
struct alias  *first_alias = NULL;
struct termios *original_term;
extern int		vacm_errno;
extern int		errno;
extern char   *optarg;
extern int    opterr;
extern int		optopt;
#ifdef HAVE_LIBREADLINE
char  hist_file[1024];
#endif

static void signal_handler(int I_signo);

int main(argc, argv)
int		argc;
char	**argv;
{
	struct timeval		to;
	fd_set						read_fds;
	int								max;
	int								rc;
#ifndef HAVE_LIBREADLINE
	struct termios		term;
#endif
	char							*netbuff;
	char							*command = NULL;
	char              *nexxus = NULL;
	char              *login = NULL;
	char              *pass = NULL;
  struct handle     *current_handle;
	int               preconnect = 0;
	int								arg_counter;

  original_term = malloc(sizeof(struct termios));
  tcgetattr(fileno(stdin), original_term);
	opterr = 0;

	signal(SIGINT, signal_handler);
	while ((rc = getopt(argc,argv,"c:p:u:vx:"))!=EOF)
		{
		if ((char)rc == 'c')
			{
				nexxus = strdup(optarg);
				preconnect = 1;
			}
		else if ((char)rc == 'p')
			{
				pass = strdup(optarg);
			}
		else if ((char)rc == 'u')
			{
				login = strdup(optarg);
			}
		else if ((char)rc == 'v')
			{
				printf("VACM %s - vash (Compiled %s %s) ", 
					VERSION,
					__DATE__,
					__TIME__);
#ifdef HAVE_LIBREADLINE
				printf("- ReadLine Enabled.\n");
#else
				printf("- ReadLine Disabled.\n");
#endif
				do_exit(0);
			}
		else if ((char)rc == 'x')
			{
			interactive = 0;
			command = strdup(optarg);
			}

		switch(optopt)
			{
			case '?':
				break;
			default:
				printf("Usage: %s <-c nexxus> <-u username> <-p password> <-x command>\n",argv[0]);
				do_exit(-1);
			}
		}
	
	/* Unset our command line */
	for (arg_counter=1; arg_counter<argc; arg_counter++)
		memset(argv[arg_counter],0,strlen(argv[arg_counter]));

	if (preconnect)
		{
		do_connect(nexxus,login,pass);
		free(nexxus);
		if (pass)
			{
			memset(pass,0,strlen(pass));
			free(pass);
			}
		if (login)
			free(login);
		}
	else
		{
		if ((login)||(pass)||(command))
			{
			printf("vash: You must connect (-c) to use -u, -p and -x options.\n");
			do_exit(-1);
			}
		}

	if (!interactive) 
		{
		/* Do our non interactive processing and exit */
		exit(handle_user_input(command));
		}
	else
		{
		/* read in our ~/.vashrc */
		char *user_dir;
		char rc_file[1024];
		user_dir = getenv("HOME");
		/* parse ~/.vashrc */
		if(*user_dir)
			{
			mode_t mask;
			FILE *Tfilep_ptr;
			
			snprintf(rc_file,sizeof(rc_file),"%s/.vashrc",user_dir);

			mask = 0x1ff;
			mask &= ~S_IRUSR;
			mask &= ~S_IWUSR;
			umask(mask);

			if ((Tfilep_ptr = fopen(rc_file, "a")))
				{
				fclose(Tfilep_ptr);
				source_cb(rc_file);
				}
			}
		}

#ifdef HAVE_LIBREADLINE
	if (interactive)
	{
	char *user_dir;
	user_dir = getenv("HOME");
	/* read in our history file */
	if(*user_dir)
		{
		snprintf(hist_file,sizeof(hist_file),"%s/.vash_history",user_dir);
		read_history(hist_file);
		}

	rl_callback_handler_install (prompt, rl_handle_user_input);
	rl_attempted_completion_function = (CPPFunction *)rl_tab_completion;
	}
#else
	if (interactive)
		{
		tcgetattr(fileno(stdin), &term);
		cfmakeraw(&term);
		term.c_iflag |=ICRNL;
		term.c_lflag |=ECHO;
		term.c_lflag |=ECHONL;
		term.c_oflag |=OPOST;
		term.c_oflag |=ONLCR;
	
		term.c_lflag |=ISIG;
		term.c_lflag |=ICANON;
		term.c_lflag |=ECHOE;
		tcsetattr(fileno(stdin), TCSANOW, &term);

		inputbuffsize = INPUT_BLOCK_SIZE;
		inputbuff = malloc(inputbuffsize);
		memset(inputbuff,0,inputbuffsize);
		}
#endif

	while(1)
		{
		max =0;
		FD_ZERO(&read_fds);
		to.tv_sec=1;
		to.tv_usec=0;
		if (interactive)
			{
			FD_SET(fileno(stdin), &read_fds);
			max =fileno(stdin);
			}
		
    current_handle = first_handle;
    while (current_handle)
			{
			int tmp_fd;
			int rc;

			if ((rc =api_nexxus_return_fd(current_handle->handle, &tmp_fd))<0)
				{
				printf("vash: Unable to lookup Nexxus FD (%s)\n",
							api_nexxus_get_error());
				do_exit(-1);
				}
			FD_SET(tmp_fd, &read_fds);
			if (tmp_fd > max)
				max = tmp_fd;

      current_handle = current_handle->next;
			}
		if ((rc = select((max+1),&read_fds, NULL, NULL, &to))<0)
			{
			/* We could receive a signal here... no worries */
			if(errno == EINTR)
				continue;
			else
				{
				perror("select");
				do_exit(-1);
				}
			}
		if (!rc)
			continue;
		if (interactive)
			{
			if (FD_ISSET(fileno(stdin), &read_fds))
				{
#ifdef HAVE_LIBREADLINE
				rl_callback_read_char ();
#else
				no_rl_callback_read_char ();
#endif
				}
			}
    current_handle = first_handle;
    while (current_handle)
			{
			int tmp_fd;
			int	len;

			if (api_nexxus_return_fd(current_handle->handle, &tmp_fd)<0)
				{
				printf("vash: Unable to lookup FD for Nexxus at %s (%s)\n",
              current_handle->name,
							api_nexxus_get_error());
				do_exit(-1);
				}
			if (FD_ISSET(tmp_fd, &read_fds))
				{
				if (api_nexxus_recv_ipc(current_handle->handle, &netbuff, &len)<0)
					{
					if (vacm_errno != API_RETURN_NEXXUS_DISCONNECT)
						{
						api_nexxus_perror("vash");
						do_exit(-1);
						}
					printf("vash: Nexxus %s disconnected unexpectedly\n",
								 current_handle->name);
					if (!interactive)
						do_exit(-1);

          api_nexxus_disconnect(current_handle->handle);
          destroy_handle(current_handle,&current_handle);
					continue;
					}
				else
					{
#ifdef HAVE_LIBREADLINE
        if(isatty(fileno(stdout)))
          {
          unsigned int x;

          /* Move to the end of the line */
          x = rl_point;
          while (x<strlen(rl_line_buffer))
            {
            printf(" ");
            x++;
            }
          for (x = 0; x < (strlen(rl_line_buffer)+strlen(prompt)); x++)
            printf("\b \b");
          printf("  %s\n",netbuff);
          rl_on_new_line();
          rl_forced_update_display();
          }
        else
					printf("%s\n",netbuff);
#else
					printf("%s\n%s",netbuff,inputbuff);
#endif
          current_handle = current_handle->next;
					free(netbuff);
					}
			  } /* handle fd set */
        else
          current_handle = current_handle->next;
			} /* handle check for loop */
		} /* while(1) */
}

#ifdef HAVE_LIBREADLINE
static void rl_handle_user_input (char *line)
{
  if (line == NULL)
		{
	 	printf("exit\n");
		do_exit(0);
		}
	else
		{
		char *dup;
		dup = strdup(line);
		handle_user_input (line);
		if (strlen(dup))
			{
			unsigned int i;
			for(i=0; i< strlen(dup); i++)
				if (!whitespace(dup[i]))
					break;
			if (!(i == strlen(dup)))
				add_history (&dup[i]);
			}
    free(line);
		free(dup);
	  }
}

static char **rl_tab_completion(char *line, int start)
{
	int i = 0;
	char **matches = NULL;

	while(rl_line_buffer[i] && whitespace (rl_line_buffer[i]))
		i++;

	/* if we are completing a command */
	if (i == start)
		matches = completion_matches(line, rl_get_full_command_matches);
	else
		{
		int j = i;
		char *cmd_buff;

		while(rl_line_buffer[j] && !whitespace (rl_line_buffer[j]))
			j++;

		cmd_buff = malloc( ++j - i );		
		snprintf(cmd_buff, (j-i), "%s", &rl_line_buffer[i]);

		while(rl_line_buffer[j] && whitespace (rl_line_buffer[j]))
			j++;

		if (j == start)
			/* Parse out a command */
			if (!strcmp("ipc", cmd_buff))
				matches = completion_matches(line, rl_get_nexxus_matches);
			else if (!strcmp("disconnect", cmd_buff))
				matches = completion_matches(line, rl_get_nexxus_matches);
			else if (!strcmp("help", cmd_buff))
				matches = completion_matches(line, rl_get_command_matches);
			else if (!strcmp("nexxus_modules", cmd_buff))
				matches = completion_matches(line, rl_get_nexxus_matches);
			else if (!strcmp("unalias",cmd_buff))
				matches = completion_matches(line, rl_get_alias_matches);
	
		free(cmd_buff);
		}
	return(matches);
}

static char *rl_get_full_command_matches(char *word, int state)
{
	static int command_index, length;
	static struct alias *current_alias;

	if (!state)
		{
		command_index=0;
		length = strlen(word);
		current_alias = first_alias;
		}

	while (commands[command_index].name)
		{
		char *name;
		name = commands[command_index].name;
		command_index++;
		if (!strncmp (name,word,length))
			return(strdup(name));
		}

	while (current_alias)
		{ 
    struct alias *tmp_alias;

    tmp_alias = current_alias;
		current_alias = tmp_alias->next;
		if (!strncmp (tmp_alias->alias,word,length))
			return(strdup(tmp_alias->alias));
		}
	return((char *)NULL);
}

static char *rl_get_command_matches(char *word, int state)
{
	static int command_index, length;

	if (!state)
		{
		command_index=0;
		length = strlen(word);
		}

	while (commands[command_index].name)
		{
		char *name;
		name = commands[command_index].name;
		command_index++;
		if (!strncmp (name,word,length))
			return(strdup(name));
		}
	return((char *)NULL);
}

static char *rl_get_alias_matches(char *word, int state)
{
	static int length;
	static struct alias *current_alias;

	if (!state)
		{
		length = strlen(word);
		current_alias = first_alias;
		}

	while (current_alias)
		{
    struct alias *tmp_alias;

    tmp_alias = current_alias;
		current_alias = current_alias->next;
		if (!strncmp (tmp_alias->alias,word,length))
			return(strdup(tmp_alias->alias));
		}
	return((char *)NULL);
}

static char *rl_get_nexxus_matches(char *nexxus, int state)
{
	static int length;
  static struct handle *current_handle;
  struct handle *tmp_handle;

	if (!state)
		{
    current_handle = first_handle;
		length = strlen(nexxus);
		}

	while (current_handle)
		{
    tmp_handle = current_handle;
    current_handle = current_handle->next;
		if (!strncmp (nexxus,tmp_handle->name,length))
      {
			return(strdup(tmp_handle->name));
      }
    }
	return((char *)NULL);
}

#else
static void no_rl_callback_read_char()
{
	int length;

	length = strlen(inputbuff);

	if (length+1 == inputbuffsize)
		{
		inputbuffsize += INPUT_BLOCK_SIZE;
		inputbuff = realloc(inputbuff, inputbuffsize);
		}
	if (read(fileno(stdin), &inputbuff[length],1)<0)
		{
		perror("read");
		do_exit(-1);
		}
	if ((inputbuff[length] == '\r') || (inputbuff[length] == '\n'))
		{
		inputbuff[length] = '\0';
		handle_user_input(strdup(inputbuff));
		free(inputbuff);
		inputbuffsize = INPUT_BLOCK_SIZE;
		inputbuff = malloc(inputbuffsize);
		memset (inputbuff, 0, inputbuffsize);
		}
}
#endif

static COMMAND *command_lookup(char *name)
{
	int i;

	for(i=0; commands[i].name; i++)
		{
		if (strcmp(name,commands[i].name)==0)
			return(&commands[i]);
		}
	return((COMMAND *)NULL);
}

static struct alias *alias_lookup(char *word)
{
  struct alias *current_ptr;
  struct alias *tmp_ptr;

  current_ptr = first_alias;
	while (current_ptr)
    {
    tmp_ptr = current_ptr;
    current_ptr = current_ptr->next;
		if (!strcmp(tmp_ptr->alias, word))
      {
			return(tmp_ptr);
      }
    }
      
	return((struct alias *)NULL);
}

static int handle_user_input(char *SZ_line)
{
	int i = 0;
	char *c;
	COMMAND *command;
	struct alias *alias;
	char *word;
	char *arg;

	while (SZ_line[i] && whitespace(SZ_line[i]))
		i++;
	SZ_line = &SZ_line[i];

	if(!strlen(SZ_line))
		return(0);

	/* Strip trailing whitespace */
	c = SZ_line + strlen(SZ_line) - 1;
	while ( c > SZ_line && whitespace (*c))
		c--;
	*++c = '\0';

	word = SZ_line;
	
	while(SZ_line[i] && !whitespace(SZ_line[i]))
		i++;

	if (SZ_line[i])
		SZ_line[i++] = '\0';

	alias = alias_lookup(word);

	if (alias)
		{
		int rc;
		char *expanded_line;
		int length = strlen(&SZ_line[i]) + strlen(alias->expanded) + 2;

		expanded_line = malloc(length);
		sprintf(expanded_line, "%s %s", alias->expanded, &SZ_line[i]);
		rc = handle_user_input(expanded_line);
		free(expanded_line);
		return(rc);
		}

	command = command_lookup(word);

	if (!command)
		{
		printf("vash: No such command. (%s)\n",word);
		return(-1);
		}
	arg = SZ_line + i;

	return((*(command->callback)) (arg));
}

static int modules_cb(char *arg)
{
	struct in_addr	addy;
	char	out_pkt[64];
	char	*bword = NULL;
	void	*handle;
	char	*in_pkt;
	int		in_len;
	int		rc;

	if (!strlen(arg))
		{
		printf("vash: nexxus_modules requires a Nexxus\n");
		return(-1);
		}

	if (api_nexxus_return_handle(&handle, arg)<0)
		{
		printf("vash: Unable to find connected Nexxus %s\n",arg);
		return(-1);
		}

	api_nexxus_return_ip(handle, &addy);
	strcpy(out_pkt,"NEXXUS:NEXXUS_MODULES");
	api_nexxus_send_ipc(handle, &out_pkt[0], (strlen(out_pkt)+1));
	while(1)
		{
		if ((rc=api_nexxus_wait_for_data(handle, &in_pkt, &in_len, 10))<0)
			{
			if (rc == API_RETURN_TIMED_OUT)
				{
				printf("vash: Nexxus %s timed out on reply\n",inet_ntoa(addy));
				return(-1);
				}
			else
				{
				perror("vash");
				do_exit(-1);
				}
			}
		if (!(bword = strtok(in_pkt, ":"))) /* NEXXUS */
			break;
		if (!(bword = strtok(NULL,":"))) /* job_id */
			break;
		if (!(bword = strtok(NULL,":"))) /* MODULE */
			break;
		if (!strcmp(bword,"MODULE"))
			{
			printf("Short Name -> %s\n",strtok(NULL,":"));
			printf("Long Name  -> %s\n",strtok(NULL,":"));
			printf("Author     -> %s\n",strtok(NULL,":"));
			printf("Desc.      -> %s\n",strtok(NULL,":"));
			printf("Version    -> %s\n",strtok(NULL,":"));
			printf("IPC Tag    -> %s\n",strtok(NULL,":"));
			printf("State      -> %s\n",strtok(NULL,":"));
			printf("\n");
			free(in_pkt);
			}
		else if (!strcmp(bword,"JOB_COMPLETED"))
			{
			free(in_pkt);
			break;
			}
		else if (!strcmp(bword,"JOB_STARTED"))
			{
			free(in_pkt);
			continue;
			}
		else
			{
			printf("Unknown nexxus status tag %s\n",bword);
			free(in_pkt);
			return(-1);
			}
		}
	printf("Nexxus Status Summary for %s completed\n",inet_ntoa(addy));
	return(0);
}

static int ipc_cb(char *arg)
{
  char                *nexxus;
	char								*bword;
	char 								*saved_msg;
	struct in_addr			addy;
	int									distance;
	void								*handle;
  int                 i=0;

	saved_msg = (char *) malloc ((strlen(arg)+1));
	strcpy(saved_msg, arg);

	if(!(nexxus = strtok(arg, " ")))
		{
		printf("vash: ipc requires a nexxus and a message to send\n");
		free (saved_msg);
		return (-1);
		}

	/* Check to make sure we're connected to the nexxus requested */
	if (api_nexxus_return_handle(&handle, nexxus)<0)
		{
		printf("vash: Unable to find connected Nexxus %s\n",nexxus);
		free(saved_msg);
		return(-1);
		}
	api_nexxus_return_ip(handle, &addy);
	/* Send the message */
	if(!(bword = strtok(NULL,":")))
		{
		printf("vash: ipc requires a nexxus and a message to send\n");
		free(saved_msg);
		return(-1);
		}
  while (whitespace(bword[i]))
    i++;
	distance = (bword - arg + i);
	api_nexxus_send_ipc(handle,
											&saved_msg[distance],
											(strlen(&saved_msg[distance])+1));
	free(saved_msg);
  if (!interactive)
    while(1)
      {
      int in_len;
      char *in_pkt;
      int rc;

      if ((rc=api_nexxus_wait_for_data(handle, &in_pkt, &in_len, 50))<0)
        {
        if (rc == API_RETURN_TIMED_OUT)
          {
          printf("vash: Nexxus %s timed out waiting for IPC response\n",
            inet_ntoa(addy));
          return(-1);
          }
        else if (rc == API_RETURN_NEXXUS_DISCONNECT)
          {
          api_nexxus_return_ip(handle, &addy);
          printf("vash: Nexxus %s disconnected unexpectedly\n",
            inet_ntoa(addy));
          disconnect_cb(nexxus);
          return(-1);
          }
        else
          {
          perror("vash");
          do_exit(-1);
          }
        }
      printf("%s\n",in_pkt);
      if (!(bword = strtok(in_pkt, ":"))) /* NEXXUS */
        {
        free(in_pkt);
        return(-1);
        }
      if (!(bword = strtok(NULL,":"))) /* job_id */
        {
        free(in_pkt);
        return(-1);
        }
      if (!(bword = strtok(NULL,":"))) /* MODULE */
        {
        free(in_pkt);
        return(-1);
        }
      if (!strcmp(bword,"JOB_STARTED"))
        {
				job_counter++;
        free(in_pkt);
        }
      else if (!strcmp(bword,"JOB_COMPLETED"))
        {
				job_counter--;
        free(in_pkt);
				if (!job_counter)
        	break;
        }
      else if (!strcmp(bword,"JOB_ERROR"))
        {
				job_counter--;
        free(in_pkt);
				if (!job_counter)
        	break;
        }
      else if (!strcmp(bword,"MODULE_NOT_FOUND"))
        {
        free(in_pkt);
        return(-1);
        }
			else
				{
				free(in_pkt);
				}
      }
      
	return(0);
}

static int connect_cb(char *arg)
{
	char *nexxus;
	char *login = NULL;
	char *pass = NULL;
  struct handle *current_handle;
	void *tmp_handle;

  current_handle = first_handle;

	if(!(nexxus = strtok(arg, " ")))
		{
		int nexxus_cnt = 0;
  	while (current_handle)
    	{
			struct in_addr	addy;

			api_nexxus_return_ip(current_handle->handle, &addy);
			printf("Nexxus %s at %s\n",
				current_handle->name,
				inet_ntoa(addy));
			nexxus_cnt++;
    	current_handle = current_handle->next;
			}
		printf("Total of %d Current Connections\n",nexxus_cnt);
		return(0);
		}
	login = strtok(NULL, " ");
	pass = strtok(NULL, " ");

	if(!(api_nexxus_return_handle(&tmp_handle,nexxus)==API_RETURN_LOOKUP_FAILED)) 
		{
		printf("vash: Already connected to Nexxus at %s\n",nexxus);
		return(-1);
		}

	return(do_connect(nexxus,login,pass));
}

static int do_connect(char *nexxus, char *login, char *pass)
{
	int rc;
	char *in_pkt;
	int  in_len;
	char p_buff[64];
	char l_buff[64];
	void *handle_ptr;
	struct handle *current_handle;
	struct hostent *host;
		
	if(!(host = gethostbyname(nexxus)))
		{
		printf("vash: Failed to look up address for host %s\n",nexxus);
		return(-1);
		}
	else
		{
		struct in_addr inaddr;
		struct timeval to;

		to.tv_sec=5;
		to.tv_usec=0;

		memcpy(&inaddr, host->h_addr_list[0], sizeof(struct in_addr));
		if (api_nexxus_ping(&inaddr,&to))
			{
			printf("vash: Unable to ping nexxus at %s\n",nexxus);
			return(-1);
			}
		}

	if (!login)
		{
		printf("%s login: ",nexxus);
		fflush(stdout);
		fgets(l_buff,sizeof(l_buff),stdin);
		l_buff[strlen(l_buff)-1] = 0;
		login = l_buff;
		}
	if (!pass)
		{
		unsigned int i = 0;
		struct termios old_term;
		struct termios new_term;

		printf("Password: ");
		fflush (stdout);
		tcgetattr(fileno(stdin), &old_term);
		cfmakeraw(&new_term);
		tcsetattr(fileno(stdin), TCSANOW, &new_term);
		while(i < sizeof(p_buff))
			{
			if (read(fileno(stdin), &p_buff[i],1)<0)
				{
				perror("read");
				do_exit(-1);
				}
			if ((p_buff[i] == '\r') || (p_buff[i] == '\n'))
				{
				p_buff[i] = 0;
				break;
				}
			else if ((p_buff[i] == '\b') || (p_buff[i] == 0x7f))
				{
				if(i)
					{
					printf("\b \b");
					i--;
					}
				}
			else if (!(iscntrl(p_buff[i])))
				{
				printf("*");
				i++;
				}
			fflush(stdout);
			}
		pass = p_buff;
		tcsetattr(fileno(stdin), TCSANOW, &old_term);
		printf("\n");	
		}

	if (api_nexxus_connect(nexxus,login,pass,&handle_ptr)<0)
		{
		memset(p_buff,0,sizeof(p_buff));
		printf("vash: Unable to connect to nexxus at %s (%s)\n",
			nexxus,
			api_nexxus_get_error());
		return(-1);
		}

	memset(p_buff,0,sizeof(p_buff));

	/* Wait for our Connect message */
	if ((rc=api_nexxus_wait_for_data(
      handle_ptr, 
      &in_pkt, 
      &in_len, 
      10))<0)
		{
		if (rc == API_RETURN_TIMED_OUT)
			{
			printf("vash: Nexxus %s timed out waiting for connect message\n",nexxus);
			return(-1);
			}
		else
			{
			perror("vash");
			do_exit(-1);
			}
		}

  if (strcmp(in_pkt, "NEXXUS_READY"))
    {
    printf("vash: Waiting for NEXXUS_READY, received %s, disconnecting\n",
      in_pkt);
    api_nexxus_disconnect(handle_ptr);
    free(in_pkt);
    return(-1);
    }

  current_handle = create_handle();
  current_handle->name = strdup(nexxus);
  current_handle->handle = handle_ptr;
  
	if (interactive)
		printf("%s\n",in_pkt);
	free(in_pkt);
	return(0);
}

static int exit_cb(char *arg)
{
	if (arg)
		if (strlen(arg))
			{
			printf("vash: Exit doesn't accept commandline arguments\n");
			return(-1);
			}
	return(do_exit(0));
}

static int do_exit(int exitval)
{
#ifdef HAVE_LIBREADLINE
	if (interactive)
		{
		if(strlen(hist_file))
			{
			int size = 100;
			char *hist_size;

			write_history(hist_file);

			hist_size = getenv("HISTSIZE");
			if(hist_size)
				sscanf(hist_size,"%d",&size);
	
			history_truncate_file(hist_file,size);
			}
		}
#endif

  if(interactive)
    tcsetattr(fileno(stdin),TCSANOW,original_term);
  free(original_term);
	exit(exitval);
}

static int disconnect_cb(char *arg)
{
  struct handle *current_handle;
	char *bword;
	void *handle;

	if (!(bword = strtok(arg, " ")))
		{
		printf("vash: Disconnect requires a Nexxus to disconnect from\n");
		return(-1);
		}
	if (api_nexxus_return_handle(&handle, bword)<0)
		{
		printf("vacm: Not connected to Nexxus at %s.\n", bword);
		return(-1);
		}

  current_handle = first_handle;
  while (current_handle->handle != handle)
    current_handle = current_handle->next;

  api_nexxus_disconnect(current_handle->handle);
  destroy_handle(current_handle,NULL);

	return(0);
}

static struct alias *create_alias()
{
  struct alias *new;
  struct alias *current_alias;

  new = (struct alias *) malloc(sizeof(struct alias));
  new->next = NULL;

  if (!first_alias)
    first_alias = new;
  else
    {
    current_alias = first_alias;
    while (current_alias->next)
      current_alias = current_alias->next;
    current_alias->next = new;
    }

  return(new);
}

static void destroy_alias(struct alias *ptr, struct alias **scan_ptr)
{
  struct alias *current_alias;

  if (ptr == first_alias)
    first_alias = first_alias->next;
  else
    {
    current_alias = first_alias;
    while (current_alias->next != ptr)
      current_alias = current_alias->next;
    current_alias->next = ptr->next;
    } 

  if(scan_ptr)
    *scan_ptr = ptr->next;

  free(ptr);    
}

static struct handle *create_handle()
{
  struct handle *new;
  struct handle *current_handle;

  new = (struct handle *) malloc(sizeof(struct handle));
  new->name = NULL;
  new->next = NULL;

  if (!first_handle)
    first_handle = new;
  else
    {
    current_handle = first_handle;
    while (current_handle->next)
      current_handle = current_handle->next;
    current_handle->next = new;
    }

  return(new);
}

static void destroy_handle( struct handle *ptr, struct handle **scan_ptr)
{
  struct handle *current_handle;

  if (*ptr->name)
    free(ptr->name);

  if (ptr == first_handle)
    first_handle = first_handle->next;
  else
    {
    current_handle = first_handle;
    while (current_handle->next != ptr)
      current_handle = current_handle->next;
    current_handle->next = ptr->next;
    }
  if (scan_ptr)
    *scan_ptr = ptr->next;

  free(ptr);    
}

static int help_cb(char *arg)
{
	int i;
	if (strlen(arg))
		{
		COMMAND *comm;
		if((comm = command_lookup(arg)))
			printf("%s - %s\n",comm->name,comm->help);
		else if (alias_lookup(arg))
			{
			printf("%s - no help is available for aliases\n",arg);
			return(-1);
			}
		else 
			{
			printf("%s - no help available on this subject\n",arg);
			return(-1);
			}
		}
	else 
		{
		char *buff;
		struct winsize wsize;

		/* Fetch our window size */
		if(ioctl(fileno(stdin),TIOCGWINSZ,(char *)&wsize)<0)
			{
			perror("ioctl");
			wsize.ws_col = 80;
			}

		buff = malloc(wsize.ws_col + 1);
		memset(buff,0,(wsize.ws_col + 1));

		printf("Help is available for the following commands:\n\n");
		for (i = 0; commands[i].name; i++)
			{
				if(((wsize.ws_col+1)-strlen(buff)) > strlen(commands[i].name))
					{
				  strcat(buff,commands[i].name);
					if (commands[i + 1].name)
						strncat(buff," ",(wsize.ws_col+1));
					}
				else
					{
					printf("%s\n",buff);
					snprintf(buff,(wsize.ws_col+1),"%s ",commands[i].name);
					}
			}
		printf("%s\n",buff);
		free(buff);
		}
	return(0);
}

static int source_cb(char *arg)
{
	int fd;
	struct stat finfo;
	extern int errno;
	char *data;
	char *line_buff;
	int rc;
	int i = 0;

	if(!strlen(arg))
		{
		printf("vash: source requires a filename to read from\n");
		return(-1);
		}

	/* Strip leading whitespace */
	while((arg[i]) && whitespace(arg[i]))
		i++;
	arg = &arg[i];

	/* Scan through the argument */
	i = 0;
	while((arg[i]) && !whitespace(arg[i]))
		i++;

#ifdef HAVE_LIBREADLINE
	arg = tilde_expand(arg);
#endif

	if ((fd = open (arg, O_RDONLY))<0)
		{
		printf("vash: Error opening file %s (%s)\n",arg,strerror(errno));
		return(-1);
		}
	
	rc = fstat(fd,&finfo);
	if (rc<0)
		{
		printf("vash: Error stating file %s (%s)\n",arg,strerror(rc));
		return(-1);
		}

	data = malloc((int)finfo.st_size +1);
	if (!data)
		{
		printf("vash: Failed to allocate memory to source %s\n",arg);
		return(-1);
		}
	read(fd, data, finfo.st_size);
	data[finfo.st_size] = '\0';
	close (fd);

	/* make sure we're not reading a binary */
	for (i = 0; i < (int)finfo.st_size; i++)
		{
		if (data[i] == '\n')
			break;
		if (!isspace(data[i]) && !isprint(data[i]))
			{
			free(data);
			printf("vash: File %s appears to be a binary\n",arg);
			return(-1);
			}
		}

	/* Everything looks OK */
	line_buff = &data[0];

	for (i = 0; i < (int)finfo.st_size; i++)
		{
		if((data[i] == '\n') || (data[i] == '\r'))
			{
			char *buff;
			int buff_len;

			buff_len = &data[i] - line_buff;
			if (buff_len)
				{
				buff = malloc(++buff_len);
				snprintf(buff, buff_len, "%s", line_buff);
				handle_user_input(buff);
				free(buff);
				}
			line_buff = &data[i+1];
			}
		}

	free(data);
	return(0);
}

static int alias_cb(char *arg)
{
	int i = 0;
  struct alias *current_ptr;
  char *alias;
	char *expanded;

  current_ptr = first_alias;

	if (!strlen(arg))
		{
    while (current_ptr)
      {
			printf("alias %s = \"%s\"\n",current_ptr->alias,current_ptr->expanded);
      current_ptr = current_ptr->next;
      }
		return(0);
		}

	while(arg[i] && !whitespace(arg[i]) && (!(arg[i] == '=')))
		i++;

  alias = malloc(i+1);
	snprintf(alias,(i+1),"%s",arg);

	while(arg[i] && whitespace (arg[i]))
		i++;

	if (!(arg[i] == '='))
		{
		printf("vash: Usage is alias <alias_name> = \"connect localhost\"\n"); 
		free(alias);
		return(-1);
		}
	i++;

	while(arg[i] && whitespace (arg[i]))
		i++;

	if (arg[i] == '"')
		{
		char *tmp;

		i++;
		tmp = &arg[i];
		while (arg[i] && (arg[i] != '"'))
			i++;

		if (arg[i] == '"')
			arg[i] = '\0';
		else
			{
			printf("vash: Unterminated quotes in command alias.\n");
			free(alias);
			return(-1);
			}
		expanded = strdup(tmp);
		}
	else
		{
		int start = i;
		while((arg[i]) && !whitespace(arg[i]))
			i++;

		expanded = malloc( i - start + 1);
		snprintf(expanded,(i - start + 1), "%s", &arg[start]);
		}

	if (!(strcmp(alias,expanded)))
		{
		printf("vash: Unable to alias %s to itself\n",alias);
		free(expanded);
		free(alias);
		return(-1);
		}

	if (command_lookup(alias))
		{
		printf("vash: Unable to use command name %s as an alias\n",alias);
		free(expanded);
		free(alias);
		return(-1);
		}
	else
		{
		/* make sure we are aliasing a command */
		char *buff;
		int start;
		i=0;
		while((expanded[i]) && whitespace(expanded[i]))
			i++;
		start = i;
		while((expanded[i]) && !whitespace(expanded[i]))
			i++;
		buff = malloc ( i - start + 1);
		snprintf(buff,(i - start + 1), "%s", &expanded[start]);
    if (alias_lookup(buff))
      {
      printf("vash: Unable to alias an alias\n");
      free(alias);
      free(expanded);
      return(-1);
      }
		if (!command_lookup(buff))
			{
			printf("vash: Unable to set aliases on unknown commands\n");
			free(alias);
			free(expanded);
			return(-1);
			}
		free(buff);
		}

	/* Check for existing alias with same name I should remove */
	if (alias_lookup(alias))
		unalias_cb(alias);


	/* Everything looks good... Add the alias */
  current_ptr = create_alias();
	current_ptr->alias = alias;
	current_ptr->expanded = expanded;

	return(0);
}

static int unalias_cb(char *arg)
{
  struct alias *current_alias;

	if (!strlen(arg))
		{
		printf("vash: Unalias requires an alias to remove\n");
		return(-1);
		}

  current_alias = first_alias;

  while(current_alias)
    {
    if (!strcmp(current_alias->alias,arg))
      {
      destroy_alias(current_alias,&current_alias);
      return(-1);
      }
    else
      current_alias = current_alias->next;
    }
    printf("vash: %s is not aliased\n",arg);
    return(-1);
}

static void signal_handler(int I_signo)
{
	if (I_signo == SIGINT)
		{
		printf("\nvash: Received SIGINT... Exiting cleanly.\n");
		do_exit(0);
		}
}
